import sys

class LiftNode:
	def __init__(self, score, label):
		self.score = score
		self.label = label

def readStratumSet(filename):
	f = open(filename, 'r')
	lines = f.readlines()
	f.close()
	retSet = set()
	for line in lines:
		line = line.strip()
		if not line:
			continue
		# get rid of the period at the end
		line = line.replace('.','')
		retSet.add(line)
	return retSet

# Computes the lift AUC of a given set of probabilities and labels,
# assuming that they are all for the same stratum.
# @param probs the probabilities given to the examples
# @param labels the class labels, 1 for positive 0 for negative
# @return
def computeLiftAUC(probs, labels):
	nodeList = list()
	for i in range(len(probs)):
		nodeList.append(LiftNode(probs[i], labels[i]))
	return nodeList

# Computes the lift AUC of a given list Sorts the lists in descending order
# Uses trapezoidal rule to compute AUC Returns the area scaled to a
# fractional X-axis [0,1].
# 
# @param list
#            The List<LiftNode> data
# @return double lift AUC score
def computeLiftAUC(nodeList):
	# sort data by their descending score
	nodeList = sorted(nodeList, key=lambda x: x.score, reverse=True)

	# initialize variables
	nbPos = nbInstance = previousNbPos = previousNbInstance = sum = 0;

	# incrementaly compute lift AUC
	for i in range(len(nodeList)):
		nbInstance += 1 # one more instance
		nbPos += nodeList[i].label # assuming POS_CLASS is not changed

		# look ahead to see if end of stretch of same score
		if i < (len(nodeList) - 1) and abs(nodeList[i].score - nodeList[i+1].score) < 0.000001:
			continue

		# not same stretch, compute lift AUC for this point increment
		# use trapezoidal rule
		sum += (nbInstance - previousNbInstance) * (nbPos + previousNbPos)
		previousNbInstance = nbInstance
		previousNbPos = nbPos

	# divide by nbInstance so that x axis is [0,1]
	return sum / (2.0 * nbInstance)

def main(args):
	f = open(args[1], 'r')
	lines = f.readlines()
	f.close()
	targetNodes = list()
	otherNodes = list()
	targetSet = readStratumSet(args[2])
	otherSet = readStratumSet(args[3])
	for line in lines:
		line = line.strip()
		if not line:
			continue
		parts = line.replace('prob(','').replace(').','').split(',')
		label = POS_CLASS if parts[0] == 'pos' else NEG_CLASS
		prob = float(parts[2])
		if parts[1] in targetSet:
			targetNodes.append(LiftNode(prob,label))
		elif parts[1] in otherSet:
			otherNodes.append(LiftNode(prob,label))
		else:
			print(parts[1] + ' not in either stratum')
	# now pass those along to compute uplift
	targetLift = computeLiftAUC(targetNodes)
	otherLift = computeLiftAUC(otherNodes)
	print('Target Lift: ' + str(targetLift))
	print('Other Lift:  ' + str(otherLift))
	print('Uplift:      ' + str((targetLift - otherLift)))

POS_CLASS = 1
NEG_CLASS = 0
main(sys.argv)

#prob(pos,invasive(113432),0.28582670083164813).