import sys, subprocess, time

class SAYL_PARAMS:
	def __init__(self, fold, subfold, threshold, minpos, searchMode, uplift):
		self.fold = fold
		self.subfold = subfold
		self.threshold = threshold
		self.minpos = minpos
		self.searchMode = searchMode
		self.uplift = uplift
		
	def __str__(self):
		ret = 'fold={0} subfold={1} thresh={2} minpos={3} sm={4} uplift={5}'
		return ret.format(self.fold, self.subfold, self.threshold, self.minpos, self.searchMode, self.uplift)

def readUpliftFromFile(filename):
	f = open(filename, 'r')
	uplift = 0.0
	for line in f:
		if line.startswith('Uplift:'):
			uplift = float(line.strip().split()[1])
	f.close()
	return uplift

def main(args):
	print('Checking results...')
	# pick parameters for each fold
	for fold in range(1, NUM_FOLDS + 1):
		print('Fold {0}'.format(fold))
		# for each fold we need to select the appropriate setting
		settingScores = list()
		for t in THRESHOLDS:
			for mp in MINPOS:
				for sm in SEARCH_MODES:
					# now concatenate the subfold probs for this setting
					# use cat and a bash regex
					fmt = 'cat mammo_icv_f{0}_sf{1}_t{2}_mp{3}_{4}_{5}/probs* > {6}'
					sfStr = '*'
					tStr = str(t).replace('.', '')
					probTmp = './probs.tmp'
					callStr = fmt.format(fold, sfStr, tStr, mp, SCORE_METHOD, sm, probTmp)
					subprocess.call(callStr, shell=True)
					# now we can compute the uplift and store it
					fmt = 'python compute_uplift.py {0} {1} {2} > {3}'
					targSet = './fs_strata0/os_target.txt'
					otherSet = './fs_strata1/os_other.txt'
					upliftTmp = './uplift.tmp'
					callStr = fmt.format(probTmp, targSet, otherSet, upliftTmp)
					subprocess.call(callStr, shell=True)
					# now we can read the result from that file and store it
					uplift = readUpliftFromFile(upliftTmp)
					result = SAYL_PARAMS(fold, 0, t, mp, sm, uplift)
					print(str(result))
					settingScores.append(result)
					# clean up for the next round
					subprocess.call('rm -f {0}'.format(probTmp), shell=True)
					subprocess.call('rm -f {0}'.format(upliftTmp), shell=True)
		# we have all of our results for this fold, pick the best setting
		b = sorted(settingScores, key=lambda r: r.uplift, reverse=True)[0]
		fmt = 'BEST - Threshold={0}  MinPos={1}  SearchMode={2}  Score={3}'
		print(fmt.format(b.threshold, b.minpos, b.searchMode, b.uplift))
		print('')

NUM_FOLDS = 10
NUM_SUB_FOLDS = NUM_FOLDS - 1
THRESHOLDS = [1.01, 1.05, 1.10]
MINPOS = [6, 12]
SEARCH_MODES = ['eager', 'em']
SCORE_METHOD = 'roc'
MAX_JOBS = 8 # max number of jobs to run at once
MAX_TIME = 60 * 65 # max time to run a batch in seconds

main(sys.argv)