package sayu;
/**
   @author Jesse Davis
*/
public class BinaryCPT {

    private int[] decode;

    /**
       holds the occurance for each pair of possible parent - child combinations
    */
    private long[][] table;
    private double[][] tableLog;
    /**
       total for each row in the CPT
    */
    private double[] totals;
    /*
      occurances for each attribute Value, if this attribute has no parents
      in the bayes net
     */
    private long[] pureProb;
    private double[] pureProbLog;

    //total number of examples in the train set
    private double total;
    //the position in 
    private int thisPos;
    //the positions of the parent attributes in the feature vector
    private int[] parents;
    //the number of values this attribute can take.
    private int numvals;
    //the psuedo count used to initialize the parameter counts
    private long init_count = 1;
    //the position of the class variable in the feature vector
    private int classPos;


   /**
       @param thisPosition the position of this attribute in the feature vector
       @param parentPositions  the positions of the parent attributes in the
              feature vector
       @param classPosition the position of the class variable in the feature vector
       @param set of data for which to build the CPT on
     */
    public BinaryCPT(int thisPosition, int[] parentPositions, int classPosition, byte[][] set){

	this.classPos = classPosition;
	this.thisPos = thisPosition;
	
	numvals = 2; //attribs[thisPos].getNumValues();

	this.parents = parentPositions;

	int count;
	decode = new int[parentPositions.length];
	/*
	  Count the number of possible combinations for all of the attributes
	  that are parents of this node. 
	*/

	if (parents.length > 0){
	    count = 2; //attribs[parentPositions[0]].getNumValues();
	    decode[0] = 1;
	    for(int ix = 1; ix < parentPositions.length; ix++){
		decode[ix] = count;
		count = count * 2;//attribs[parentPositions[ix]].getNumValues();
	    }
	    table = new long[numvals][count];
	    tableLog = new double[numvals][count];
	    
	    totals = new double[count];
	    double norm = table.length * table[0].length;//count;//
	    //initialize the probabilities in the table
	    //RED FLAG - M estimates??
	    for(int ix = 0; ix < count; ix++){
		for(int jx = 0; jx < numvals; jx++){
		    table[jx][ix] = init_count;// / norm;
		    totals[ix] += table[jx][ix];
		}
	    }
	    buildCPT(set);
	    for(int thisIndex = 0; thisIndex < tableLog.length; thisIndex++){
		for(int index = 0; index < tableLog[thisIndex].length; index++){
		    tableLog[thisIndex][index] = Math.log(((double)table[thisIndex][index])/totals[index]);
		}
	    }
	}
	/*
	  This is a pure probability. no parents, so just count the occurances
	  of each value.
	*/
	else{
	    pureProb = new long[numvals];
	    pureProbLog = new double[numvals];
	    
	    //RED FLAG M estimate
	    for(int ix = 0; ix < pureProb.length; ix++){
		pureProb[ix] = init_count;// / ((double)numvals);
		total += pureProb[ix];
	    }
	    buildPureProb(set);
	    for(int ix = 0; ix < pureProbLog.length; ix++){
		pureProbLog[ix] = Math.log(((double)pureProb[ix])/total);
	    }
	}
    }//end of constructor

    private int buildIndex(byte[] vals){
	int position; //the current parent being evaluated
	int index = 0; //total index into table
	int tempIndex = 0; //index component for one attribute
   
	/*
	  go through each attribute that is a point of this node in the
	  graph and find its value.  the index into a table is found in 
	  this manner:
	  attribIndex * number of prior combinations of attribs 
	*/
	int bound = parents.length;
	for(int ix = 0; ix < bound; ix++){
	    position = parents[ix];
	    tempIndex = vals[position];
	    //build up the index
	    tempIndex = tempIndex * decode[ix];
	    index += tempIndex;
	}
	return index;
    }//end of buildIndex

    /**
       
       @return
    */
    private int[] unbuildIndex(int index){
	int[] vals = new int[decode.length];
	int remainder;
	int quo = index;
	for(int ix = (decode.length  - 1); ix >=0; ix--){
	    remainder = quo % decode[ix];
	    quo = quo / decode[ix];
	    vals[ix] = quo;
	    quo = remainder;
	}
	return vals;
    }//end unbuildIndex();

    /**
       @param train - the data to build up the probabilty
 

       this method builds a probability table for an attribute. it is a pure
       probability.  just finds the probability for each attribute value in 
       the training set.  

       note - the method doesn't actually compute the probility, it just
       counts the number of times each attribute value occurs in the 
       train set.
    */
    private void buildPureProb(byte[][] data){
	int bound = data.length;
	//build up a count for the occurance of each attribute value
	for(int ix = 0; ix < bound; ix++){
	    pureProb[data[ix][thisPos]]++;
	}//end of for loop for the count
	total += data.length;
	
    }//end of method pureProb

    /**
       @param train - the train set to build the cpt on
       
       note - the method doesn't actually compute the probilities, it just
       counts the occurances for each parent-child combination in the 
       training set

    */
    private void buildCPT(byte[][] data){
	int index; //the index for all of the parents
	int thisIndex; //the index for the value of the attribute at thisPos 
	
	/**
	   count all of the occurance for each pair of
	*/ 
	int bound = data.length;
	for(int ix = 0; ix < bound; ix++){
	    /*
	      finds the values of all of the parents and the 
	      corresponding index into the conditional prob table
	     */
	    index = buildIndex(data[ix]);
	    thisIndex = data[ix][thisPos];
	    table[thisIndex][index]++;
	    totals[index]++; 

	}//end of for loop that counts occurances
    }//end of method buildCPT

    /**
       Returns the probability for the feature vector example, for the attribute
       modelled by this CondProb.  If the attribute has no parents, then the
       value return is a pure probability, otherwise, it is a conditional 
       probability based on the values of this attributes parents

       @param example - the feature vector to be evaluated
       @param useClass
       @param out
       @return the probility/condprob for the attribute at thisPos based on
               the values in the feature vector example

    */
    public final double getProb(byte[] example, boolean useClass, int out){
	int thisIndex;//the index for the value of the attribute at thisPos 
	int index; //index into the CPT based on the parent value

	if (thisPos == classPos && useClass){
	    thisIndex = out;
	}
	else{
	    thisIndex = example[thisPos];
	}

	//return the pure probability
	if (parents.length == 0){
 	    return (((double)pureProb[thisIndex])/total);
	}
	//return the conditional probability
	else{
	    index = buildIndex(example, useClass, out);
	    return (((double)table[thisIndex][index])/totals[index]);
	}
    }//getProb(int[] example)


    /**
       Returns the probability for the feature vector example, for the attribute
       modelled by this CondProb.  If the attribute has no parents, then the
       value return is a pure probability, otherwise, it is a conditional 
       probability based on the values of this attributes parents

       @param example - the feature vector to be evaluated
       @param useClass
       @param out
       @return the probility/condprob for the attribute at thisPos based on
               the values in the feature vector example

    */
    public final double getLogProb(byte[] example, boolean useClass, int out){
	int thisIndex;//the index for the value of the attribute at thisPos 
	int index; //index into the CPT based on the parent value

	if (thisPos == classPos && useClass){
	    thisIndex = out;
	}
	else{
	    thisIndex = example[thisPos];
	}

	//return the pure probability
	if (parents.length == 0){
 	    return pureProbLog[thisIndex];
	}
	//return the conditional probability
	else{
	    index = buildIndex(example, useClass, out);
	    return tableLog[thisIndex][index];
	}
    }//getProb(int[] example)

 

    public int getFreeParameters(){
	if (parents.length == 0){
	    return pureProb.length - 1;
	}
	else{
	    return (table[0].length-1) * (table.length);
	}
    }

    public int[] getParents(){
	return parents;
    }

    public long[] getPure(){
	return pureProb;
    }
    public long[][] getTable(){
	return table;
    }

    public int getPosition(){
	return thisPos;
    }
    public boolean equals(Object o){
	BinaryCPT co = (BinaryCPT) o;
	int[] tp = co.getParents();
	if (thisPos != co.getPosition() || 
	    tp.length != parents.length){
	    System.out.println("position mismatch");
	    return false;
	}
	for(int ix = 0; ix < tp.length; ix++){
	    if (parents[ix] != tp[ix]){
		System.out.print("parents mismatch " + thisPos + " "  );

		for(int kx = 0; kx < parents.length; kx++){
		    System.out.print("(" +parents[kx] + " " + tp[kx] + ") ");
		}
		System.out.println();
		return false;
	    }
	}
	if (parents.length == 0){
	    long[] pp = co.getPure();
	    for(int ix = 0; ix < pp.length; ix++){
		if (pp[ix] != pureProb[ix]){
		    System.out.println("counts mismatch - pp");
		    return false;
		}
	    }
	}
	else{
	    long[][] pt = co.getTable();
	    for(int ix = 0; ix < pt.length; ix++){
		for(int jx = 0; jx < pt[ix].length; jx++){
		    if (pt[ix][jx] != table[ix][jx]){
			return false;
		    }
		}
	    }
	    
	}
	return true;
    }

    public void printParents(){
	System.out.print(thisPos + " |");
	for(int ix = 0; ix < parents.length; ix++){
	    System.out.print(" " + parents[ix]);
	}
	System.out.println();
    }
    public int getNumParents(){
	return parents.length;
    }

    /**
       REDFLAG COMMENT THIS SOON 
       This method builds the index into the conditional probability table
       for the attribute at this thisPos. It takes into consideration the
       all the values of the parents

       @param example
       @param attributes
       @return the index into the CPT
    */
    private int buildIndex(byte[] vals,	boolean useClass, int out){
	int position; //the current parent being evaluated
	int index = 0; //total index into table
	int tempIndex = 0; //index component for one attribute
   
	/*
	  go through each attribute that is a point of this node in the
	  graph and find its value.  the index into a table is found in 
	  this manner:
	  attribIndex * number of prior combinations of attribs 
	*/
	int bound = parents.length;
	for(int ix = 0; ix < bound; ix++){
	    position = parents[ix];
	    tempIndex = vals[position];
	   
	    if ((position == classPos) && useClass){

		tempIndex = out;
	    }
	    //build up the index
	    tempIndex = tempIndex * decode[ix];
	    index += tempIndex;

	}
	return index;
    }//end of buildIndex
    
    public void simplePrint(){
	//System.out.println();
	if (parents.length == 0){
	    for(int ix = 0; ix < pureProb.length; ix++){
		System.out.print(pureProb[ix] / ((double)total) + " ");
	    }
	}
	else{
	    for(int ix = 0; ix < table[0].length; ix++){
		for(int jx = 0; jx < table.length; jx++){
		    System.out.print(table[jx][ix] / ((double)totals[ix]) + " ");
		}
		System.out.println();
	    }
	}
	System.out.println();
    }
}
