package sayu;
import java.util.*;
public class Discretize{

    public static int debugFeature = 0;

    public static void setdebug(){
        debugFeature = 4;
    }

    public static void unsetdebug(){
        debugFeature = 0;
    }
    /**
       This method performs an aggregation to propositionalize a proposed clause. 

       @param trainPosCount the # of yes answers for each positive key
       @param trainPosTotal the total # of 
       @param trainNegCount the bitmap for the negative training examples
       @param trainNegTotal the bitmap for the negative training examples
       @param tunePosCount the bitmap for the positive tune examples
       @param tunePosTotal the bitmap for the positive tune examples
       @param tuneNegCount the bitmap for the negative tune examples
       @param tuneNegTotal the bitmap for the negative tune examples
       @param testPosCount the bitmap for the positive test examples
       @param testPosTotal the bitmap for the positive test examples
       @param testNegCount the bitmap for the negative test examples
       @param testNegTotal the bitmap for the negative test examples

       @param numValues the number of values the new feature can take on
       @param typeOfAggration how to discrete the feature (0 == Fraction)
       @return the score of the classifier incorporating the new feature.  The score is according to the <score function> that user selected
     */
    public static byte[][] generateFeature(long[] trainPosCount, long[] trainPosTotal, long[] trainNegCount, long[] trainNegTotal, long[] tunePosCount, long[] tunePosTotal, long[] tuneNegCount, long[] tuneNegTotal, long[] testPosCount, long[] testPosTotal, long[] testNegCount,   long[] testNegTotal, double[] cutOffs, int typeOfAggregation){

        //convert the 
        double[][] fractions = generateFractions(trainPosCount, trainPosTotal, trainNegCount, trainNegTotal, tunePosCount, tunePosTotal, tuneNegCount, tuneNegTotal, testPosCount, testPosTotal,  testNegCount, testNegTotal);
        if (false){
            for(int ix = 0; ix < fractions.length; ix++){
                for(int jx = 0; jx < fractions[ix].length; jx++){
                    System.out.print(Utils.round(fractions[ix][jx], 4) + " ");
                }
                System.out.println();
            }
        }
        byte[][] feats = new byte[6][];
        double[] bins = findBins(fractions[0], cutOffs.length+1, new StringBuffer());
        System.arraycopy(bins, 0, cutOffs, 0, bins.length);
        if (bins[0] == -1){
            return feats;
        }
        byte[] currentFeat = applyCut(fractions[0], bins);
        feats[0] = new byte[trainPosCount.length];
        feats[1] = new byte[trainNegCount.length];
        System.arraycopy(currentFeat, 0, feats[0], 0, feats[0].length);
        System.arraycopy(currentFeat, trainPosCount.length, feats[1], 0, feats[1].length);

        feats[2] = new byte[tunePosCount.length];
        feats[3] = new byte[tuneNegCount.length];
        currentFeat = applyCut(fractions[1], bins);
        System.arraycopy(currentFeat, 0, feats[2], 0, feats[2].length);
        System.arraycopy(currentFeat, tunePosCount.length, feats[3], 0, feats[3].length);

        feats[4] = new byte[testPosCount.length];
        feats[5] = new byte[testNegCount.length];
        currentFeat = applyCut(fractions[2], bins);
        System.arraycopy(currentFeat, 0, feats[4], 0, feats[4].length);
        System.arraycopy(currentFeat, testPosCount.length, feats[5], 0, feats[5].length);

        return feats;
    }

    /**
       This method introduces a new feature into the class.  It learns a classifier incorporating this feature, then scores the new classifier on the tune set accroding to the score function the user selected. 

       @param trainPosCount the bitmap for the positive training examples
       @param trainPosTotal the bitmap for the positive training examples
       @param trainNegCount the bitmap for the negative training examples
       @param trainNegTotal the bitmap for the negative training examples
       @param tunePosCount the bitmap for the positive tune examples
       @param tunePosTotal the bitmap for the positive tune examples
       @param tuneNegCount the bitmap for the negative tune examples
       @param tuneNegTotal the bitmap for the negative tune examples
       @param testPosCount the bitmap for the positive test examples
       @param testPosTotal the bitmap for the positive test examples
       @param testNegCount the bitmap for the negative test examples
       @param testNegTotal the bitmap for the negative test examples

       @param numValues the number of values the new feature can take on
       @param typeOfAggration how to discrete the feature (0 == Fraction)
       @return the score of the classifier incorporating the new feature.  The score is according to the <score function> that user selected
     */
    public static double[][] generateFractions(long[] trainPosCount, long[] trainPosTotal, long[] trainNegCount, long[] trainNegTotal, long[] tunePosCount, long[] tunePosTotal, long[] tuneNegCount, long[] tuneNegTotal, long[] testPosCount, long[] testPosTotal, long[] testNegCount,   long[] testNegTotal){

        double[][] feat =new double[3][];
        feat[0] = new double[trainPosCount.length + trainNegCount.length];
        feat[1] = new double[tunePosCount.length + tuneNegCount.length];
        feat[2] = new double[testPosCount.length + testNegCount.length];
        int offset = trainPosCount.length;
        //calculate the fraction for the train set
        int pos = 0;
        for(int ix = 0; ix < feat[0].length; ix++){
            if (ix < offset){
                if (trainPosTotal[ix] == 0){
                    feat[0][ix] = 0;
                }
                else{
                    feat[0][ix] = ((double)trainPosCount[ix]) / ((double)trainPosTotal[ix]);
                }
            }
            else{
                if (trainNegTotal[ix-offset] == 0){
                    feat[0][ix] = 0;
                }
                else{
                    feat[0][ix] = ((double)trainNegCount[ix - offset]) / ((double)trainNegTotal[ix-offset]);
                }
            }
        }
        //calculate the fraction for the tune set
        offset = tunePosCount.length;
        for(int ix = 0; ix < feat[1].length; ix++){
            if (ix < offset){
                if (tunePosTotal[ix] == 0 ){
                    feat[1][ix] = 0;
                }
                else{
                    feat[1][ix] = ((double)tunePosCount[ix]) / ((double)tunePosTotal[ix]);
                }
            }
            else{
                if (tuneNegTotal[ix-offset] == 0 ){
                    feat[1][ix] = 0;
                }
                else{
                    feat[1][ix] = ((double)tuneNegCount[ix - offset]) / ((double)tuneNegTotal[ix-offset]);
                }
            }
        }
        //calculate the fraction for the test set
        offset = testPosCount.length;
        for(int ix = 0; ix < feat[2].length; ix++){
            if (ix < offset){
                if (testPosTotal[ix] == 0){
                    feat[2][ix] = 0;
                }
                else{
                    feat[2][ix] = ((double)testPosCount[ix]) / ((double)testPosTotal[ix]);
                }
            }
            else{
                if (testNegTotal[ix-offset] == 0){
                    feat[2][ix] = 0;
                }
                else{
                    feat[2][ix] = ((double)testNegCount[ix - offset]) / ((double)testNegTotal[ix-offset]);
                }
            }
        }
        //return the feature
        return feat;
    }


    /**
       @param feature
       @param bins
       @param attribute
       @return
     */
    public static double[] findBins(double[] feature, int bins, StringBuffer attr){
        double[] tmpFeat = new double[feature.length];
        System.arraycopy(feature, 0, tmpFeat, 0, feature.length);
        Arrays.sort(tmpFeat);
        double[] cuts = new double[bins-1];
        double prevCut = tmpFeat[0];
        int prevCovered = 1;
        int bestDiff = 0;
        LinkedList<CutObj> allCuts = new LinkedList<CutObj>();
        for(int ix = 1; ix < tmpFeat.length; ix++){
            if (prevCut < tmpFeat[ix]){
                //System.out.println("cut " + tmpFeat[ix] + " " + ix);
                allCuts.add(new CutObj( tmpFeat[ix], ix ));
                prevCut = tmpFeat[ix];

            }
        }
        CutObj[] finalCuts = new CutObj[allCuts.size()];
        for(int ix =  0; ix < finalCuts.length; ix++){
            finalCuts[ix] = allCuts.removeFirst();
        }
        //sort the cuts in ascending order
        Arrays.sort(finalCuts);
        /*
	  either the all attributes have the same value so it is worthless to 
	  score

	  or 

	  it is a binary feature (i.e. existential aggregation) and it will
	  already be has been/will be scored as such

         */
        if (finalCuts.length == 0 || finalCuts.length == 1){
            for(int ix = 0; ix < cuts.length; ix++){
                cuts[ix] = -1;
            }

        }
        else{
            //try all pairs of cuts
            for(int ix = 0; ix < (finalCuts.length-1); ix++){
                int pos1 = finalCuts[ix].getPos();
                for(int jx = (ix+1); jx < finalCuts.length; jx++){
                    int pos2 = finalCuts[jx].getPos();
                    //quantity to minimize
                    int tmpDiff = CutObj.calcDiff(pos1, pos2-pos1, tmpFeat.length-pos2);
                    if ((ix == 0 && jx == 1) || tmpDiff < bestDiff){
                        bestDiff = tmpDiff;
                        cuts[0] = finalCuts[ix].getCut();
                        cuts[1] = finalCuts[jx].getCut();
                    } 
                }
            }
        }
        if (debugFeature >= 3){
            for(int ix = 0; ix < cuts.length; ix++){
                System.out.print("cut = " + cuts[ix] +  " ");
            }
            System.out.println();
        }
        for(int ix = 0; ix < cuts.length; ix++){
            attr.append(" lt" + cuts[ix]);
        }
        attr.append(" gte" + cuts[cuts.length-1]);
        return cuts;

    }//end of method bin


    /**

     */
    public static byte[] applyCut(double[] feature, double[] cuts){
        byte[] newFeat = new byte[feature.length];
        boolean done;
        for(int ix = 0; ix < newFeat.length; ix++){
            done = false;
            for(int jx = 0; jx < cuts.length && !done; jx++){
                if (feature[ix] < cuts[jx]){
                    newFeat[ix] = (byte)jx;
                    done = true;
                }
            }
            if (!done){
                newFeat[ix] = (byte)cuts.length;
            }
        }
        return newFeat;
    }//end of method applyCut

    /**
       This method bins the average of binary attribute into three bins.
       the bins are determined as follows:
       0 <= x < 1/3    ==> 0 
       1/3 <= x < 2/3  ==> 1 
       2/3 <= x <= 1   ==> 2 
       @param feature is the average for each example
       @return the feature discretized as describe above
     */
    public static int[] binBinary(double[] feature, StringBuffer attr){
        int[] newFeat = new int[feature.length];
        double third = 1.0/3.0;
        double twoThird = 2.0/3.0;
        attr.append(" lt" + third + " lt" + twoThird + " gte" + twoThird);
        for(int ix = 0; ix < feature.length; ix++){
            if (feature[ix] <= third){
                newFeat[ix] = 0;
            }
            else if (feature[ix] <= twoThird){
                newFeat[ix] = 1;
            }
            else{
                newFeat[ix] = 2;
            }
        }
        return newFeat;
    }//end of method binBinaray

    public static void getRandom(){
        for(int ix = 0; ix < 20; ix++){
            System.out.println(Math.random());
        }
    }

    public static void main(String[] args){
        //getRandom();
        long[] t1a = new long[5];
        long[] t1b = new long[5];
        long[] t2a = new long[5];
        long[] t2b = new long[5];
        long[] t3a = new long[5];
        long[] t3b = new long[5];

        long[] t4a = new long[5];
        long[] t4b = new long[5];
        long[] t5a = new long[5];
        long[] t5b = new long[5];
        long[] t6a = new long[5];
        long[] t6b = new long[5];

        for(int ix = 0; ix < 5; ix++){
            t1a[ix] = (long)(3 * Math.random());
            t2a[ix] = (long)(3 * Math.random());
            t3a[ix] = (long)(3 * Math.random());
            t4a[ix] = (long)(3 * Math.random());
            t5a[ix] = (long)(3 * Math.random());
            t6a[ix] = (long)(3 * Math.random());

            t1b[ix] = ((long)(3 * Math.random()) + t1a[ix]);
            t2b[ix] = ((long)(3 * Math.random()) + t2a[ix]);
            t3b[ix] = ((long)(3 * Math.random()) + t3a[ix]);
            t4b[ix] = ((long)(3 * Math.random()) + t4a[ix]);
            t5b[ix] = ((long)(3 * Math.random()) + t5a[ix]);
            t6b[ix] = ((long)(3 * Math.random()) + t6a[ix]);
        }
        //	for(int ix = 0; ix < 6; ix++){
        //}
        for(int jx = 0; jx < t1a.length; jx++){
            System.out.print(t1a[jx] + "/" + t1b[jx] + " ");
        }
        System.out.println();
        for(int jx = 0; jx < t2a.length; jx++){
            System.out.print(t2a[jx] + "/" + t2b[jx] + " ");
        }
        System.out.println();
        for(int jx = 0; jx < t3a.length; jx++){
            System.out.print(t3a[jx] + "/" + t3b[jx] + " ");
        }
        System.out.println();

        for(int jx = 0; jx < t4a.length; jx++){
            System.out.print(t4a[jx] + "/" + t4b[jx] + " ");
        }
        System.out.println();
        for(int jx = 0; jx < t5a.length; jx++){
            System.out.print(t5a[jx] + "/" + t5b[jx] + " ");
        }
        System.out.println();
        for(int jx = 0; jx < t6a.length; jx++){
            System.out.print(t6a[jx] + "/" + t6b[jx] + " ");
        }
        System.out.println();
        System.out.println("percent");
        double[] bins = new double[2];
        byte[][] nf =  Discretize.generateFeature(t1a,t1b,t2a,t2b,t3a,t3b,t4a,t4b,t5a,t5b,t6a,t6b, bins, 0);
        System.out.println("bins");
        for(int ix = 0; ix < bins.length; ix++){
            System.out.print(bins[ix] + " ");
        }
        System.out.println();
        System.out.println("discrete feature");
        for(int ix = 0; ix < nf.length; ix++){
            for(int jx = 0; jx < nf[ix].length; jx++){
                System.out.print(nf[ix][jx] + " ");
            }
            System.out.println();
        }
    }
}
