package sayu;
/**
   @author Jesse Davis 
   (c) June 2003 
   
   A class to represent an edge in a graph.  It assumes that the vertices
   in the graph are labeled with integers.  The edge can be directed.
   Additionally the edge can be weighted. The default weight is zero.

*/
public class Edge implements Comparable{

    /**
       The start vertex of the graph
     */
    private int start;

    /**
       The end vertex of the graph
    */
    private int end;

    /**
       Whether the edge is directed or not
    */
    private boolean directed;


    /**
       holds whether this edge has been visited
    */

    private boolean visited;

    /**
       The edge weight
    */
    private double weight = 0;

    /**
       
     */
    double[][][] probTable;

    /**
       @param start is the vertex the edge starts from (only matters if the edge is directed).
       @param end is the vertex the edge ends (only matters if the edge is directed).
       @param directed true iff the edge is a directed edge from start -> end. 
    */
    public Edge(int start, int end, boolean directed){
	this.start = start;
	this.end = end;
	this.directed = directed;
	visited = false;
    }

    /**
       @param start is the vertex the edge starts from (only matters if the edge is directed).
       @param end is the vertex the edge ends (only matters if the edge is directed).
       @param directed true iff the edge is a directed edge from start -> end. 
       @param weight the weight of the edge.
    */
    
    public Edge(int start, int end, boolean directed, double weight){
	this(start, end, directed);
	this.weight = weight;
    }

    /**
       @param start is the vertex the edge starts from (only matters if the edge is directed).
       @param end is the vertex the edge ends (only matters if the edge is directed).
       @param weight the weight of the edge.
       @param probTable
    */
    
    public Edge(int start, int end, double weight, double[][][] probTable ){
	this.start = start;
	this.end = end;
	this.directed = false;
	visited = false;
	this.weight = weight;
	this.probTable = probTable;
    }

    public void updateVertex(int oldVertex, int newVertex){
	
	if (start == oldVertex && newVertex < end){
	    start = newVertex;
	}
	else if (end == oldVertex && newVertex > start){
	    end = newVertex;
	}
	else{
	    double[][][] newProbTable = new double[probTable[0].length][probTable.length][probTable[0][0].length];
	    for(int ix = 0; ix < probTable[0].length; ix++){
		for(int jx = 0; jx < probTable.length; jx++){
		    for(int kx = 0; kx < probTable[ix][jx].length; kx++){
			newProbTable[jx][ix][kx] = probTable[ix][jx][kx];
		    }
		}
	    }
	    probTable = newProbTable;
	    if (oldVertex == start){
		start = newVertex;
	    }
	    else{
		end = newVertex;
	    }
	    if (end < start){
		int tmp = end;
		end = start;
		start = tmp;
	    }
	}
    }

    /**
       @return the 
     */
    public double[][][] getProbTable(){
	return probTable;
    }


    /**
       @return the start vertex
     */
    public int getStart(){
	return start;
    }

    /**
       @return the end vertex
    */
    public int getEnd(){
	return end;
    }

    /**
       @return true iff this is a directed edge
    */
    public boolean isDirected(){
	return directed;
    }

    /**
       @return whether the edge has been visited
    */
    public boolean isVisited(){
	return visited;
    }

    /**
       sets the edge to having been visited
    */
    public void visit(){
	visited = true;
    }

    /**
       resets the visited flag to false
    */
    public void resetVisit(){
	visited = false;
    }

    /**
       @param obj an edge object to compare against this edge.
       @return true iff this.start == obj.getStart() && this.end obj.getEnd().
    */
    public boolean equals(Object obj){
	Edge e = (Edge) obj;
	if ((e.getStart() == start)
	    && (e.getEnd() == end)){
	    return true;
	}
	else{
	    return false;
	}

    }//end of method equals

    /**
       @return the weight of the edge
    */
    public double getWeight(){
	return weight;
    }//end of method weight
    
    public int compareTo(Object o){
	Edge e1 = (Edge) o;
	double w = e1.getWeight();
	
	if (weight < w){
	    return -1;
	}
	else if (weight > w){
	    return 1;
	}
	else{
	    int max1 = Math.max(start, end);
	    int max2 = Math.max(e1.getStart(), e1.getEnd());
	    if (max1 > max2){
		return -1;
	    }
	    else if (max1 < max2){
		return 1;
	    }
	    else{
		return 0;
	    }
	}

    }//end of method compareTo

    public void makeDirected(int parent){
	directed = true;
	
	if (parent == end){
	    end = start;
	    start = parent;
	}
	else if (start != parent){
	    System.err.println("Bad");
	}
    }

    public String toString(){
	String rt = end + " " + weight;
	if (end < 10){
	    rt = "  " + rt;
	}
	else if (end < 100){
	    rt = " " + rt;
	}
	if (start < 10)
	    rt = "  " + start + rt;
	else if (start < 100){
	    rt = " " + start + rt;
	}
	else{
	    rt = start + rt;
	}
	return rt;
    }

    public void setEnd(int end){
	this.end = end;
    }

    public void setStart(int start){
	this.start = start;
    }
}//end of class
