

package sayu;

public class GeneralFunctions{

    /**
       @param src the int array
       @param startPos the position to start copying from in the source array
       @param dest the destination byte array
       @param destPos the position to start copying into in the destination array
       @param len how many entries to copy
     */
    public static void intToByte(int[] src, int startPos, byte[] dest, int destPos, int len){
        int curPos = destPos;
        for(int ix = startPos; ix < len; ix++){
            dest[curPos] = (byte)src[ix];
            curPos++;
        }
    }
    /**
       calculates the conditional probability table for attribute at
       position pos in the feature.  The parent is supposed to the
       class variable.

       @param train - trainset
       @param pos - position of attribute in feature vector
       @param classPos - position of class variable in feature vector
       @param numVals - number of values of attribute
       @param numClass - number of classes
     */
    public static double[][] condCountOneAttrib(int[][] train, int pos, 
            int classPos, int numVals, 
            int numClass){
        //	System.out.println("Maxtan cond count");
        long[][] counts = new long[numVals][numClass];
        //for(int ix = 0; ix < counts.length; 
        for(int ix = 0; ix < counts.length;ix++){
            for(int jx = 0; jx < counts[ix].length; jx++){
                counts[ix][jx] = 1;
            }
        }
        //calculate the counts
        for(int ix = 0; ix < train.length; ix++){
            if (train[ix][pos] >= counts.length){
                System.out.println(pos + " " + numVals + " " + counts.length);
            }
            counts[train[ix][pos]][train[ix][classPos]]++;
        }
        double[][] probs = new double[counts.length][counts[0].length];
        for(int ix = 0; ix < counts[0].length;ix++){
            long total = 0;
            //get totals
            for(int jx = 0; jx < counts.length; jx++){
                total += counts[jx][ix];
            }
            //calculate the conditional probabilities
            for(int jx = 0; jx < counts.length; jx++){
                //System.out.print(counts[jx][ix] + " ");c
                probs[jx][ix] = (double)counts[jx][ix] / (double)total;
            }
            //  System.out.println(" / " + total);
        }//end of ix loop for getting totals/probs

        return probs;
    }//end of method condCountOneAttrib


    /**
       @param attributes information about each feature 
       @param trainSet the training data
       @param outputPosition the position in the feature vector of the class variable
       @return the cpt for each node in the network
     */
    public static CondProb[] buildNaiveBayes(Datapairs[] attributes, int[][] trainSet, int outputPosition){
        int[] parents = new int[1];
        parents[0] = outputPosition;
        CondProb[] cpts =new CondProb[trainSet[0].length];
        //System.out.println(trainSe
        for(int ix = 0; ix < trainSet[0].length; ix++){
            if (ix == outputPosition){
                cpts[ix] = new CondProb(ix, new int[0], outputPosition,trainSet, attributes);
            }
            else{
                cpts[ix] = new CondProb(ix, parents, outputPosition, trainSet,
                        attributes);
            }
        } 
        return cpts;
    }

    /**
       @param dataSet the data to find the log odds ratio for
       @param cpts the cpts for the network
       @return an array where each entry is log(Prob(Neg)) - log(Prob(Pos))
     */
    public static double[] logOdds(int[][] dataSet, CondProb[] cpts){
        double probPos;
        double probNeg;
        double[] odds = new double[dataSet.length];
        for(int ix = 0; ix < dataSet.length; ix++){
            for(int jx = 0; jx < dataSet[ix].length; jx++){
                probPos = cpts[jx].getLogProb(dataSet[ix],true, 1);
                probNeg = cpts[jx].getLogProb(dataSet[ix],true,0);
                odds[ix] += (probNeg - probPos);
            }
        }
        return odds;
    }

    /**
       @param dataSets
       @param useAggs
       @return 
     */
    public static Data readData(int[] dataSets, int useAggs){
        Data data;
        int test = -1;
        for(int ix = 0; ix < dataSets.length; ix++){
            //System.out.println("fold " + ix + " = " + dataSets[ix]);
        }
        for(int ix =0; ix < dataSets.length; ix++){
            if (dataSets[ix] == 2){
                test = ix;
            }
        }
        if (useAggs == 5){
            System.out.println("Creating initial structure with 50 WARMR Features");
            data = new Data("/pdata/jans/warmr/arff/SAYU/View/", "train50fold" + test, "tune50fold" + test, "test50fold" + test, ".data",".key", "fs50fold" + test + ".names", ",", true, true);
        }
        else if (useAggs == 6){
            System.out.println("Creating initial structure with 100 WARMR Features");
            data = new Data("/pdata/jans/warmr/arff/SAYU/View/", "train100fold" + test, "tune100fold" + test, "test100fold" + test, ".data",".key", "fs100fold" + test + ".names", ",", true, true);
        }
        else if (useAggs == 7){
            System.out.println("Creating initial structure with 150 WARMR Features");
            data = new Data("/pdata/jans/warmr/arff/SAYU/View/", "train150fold" + test, "tune150fold" + test, "test150fold" + test, ".data",".key", "fs150fold" + test + ".names", ",", true, true);
        }
        else if (useAggs == 9){
            System.out.println("Creating initial structure with WARMR features");
            data = new Data("/pdata/jans/warmr/uai/tan", ".test", dataSets, test, "/pdata/jans/warmr/uai/tan0.names", ",", true, true);

        }
        else if (useAggs == 1){
            System.out.println("Creating initial structure with aggregate features");

            data = new Data("/u/jdavis/Yap/ILP/AlephUW/dev/data/old/level2/tmp/agg/", "trainAgg" + test, "tuneAgg" + test, "testAgg" + test, ".data",".key", "fold" + test + ".names", ",", true, true);

            // data = new Data("/u/jdavis/Yap/ILP/AlephUW/dev/data/old/level3/testAgg", ".data", dataSets, test, "/u/jdavis/Yap/ILP/AlephUW/dev/data/old/level3/agg.names", ",",  true, true);
        }
        else if (useAggs == 0){
            System.out.println("Creating initial structure only radiologist features");
            data = new Data("/u/jdavis/Yap/ILP/AlephUW/dev/data/old/level2/test", ".data", dataSets, test, "/u/jdavis/Yap/ILP/AlephUW/dev/data/old/level2/bc-expert.names", ",", true, true);
        }
        else if (useAggs == 2){
            for(int xx = 0; xx < dataSets.length; xx++){
                System.out.print(dataSets[xx] + " ");

            }
            System.out.println();
            System.out.println("Creating initial structure only radiologist features with matched data");
            data = new Data("/u/mammobayes/DATA/matched outcome/jesse's/notreport/test", ".data", dataSets, test, "/u/mammobayes/DATA/matched outcome/jesse's/notreport/bc-new.names", ",", true, true);

        }
        else if (useAggs == 3){
            //data = null;
            //Flag - this should really 
            System.out.println("Creating initial structure with radiologist features on matched data (includes birads)");
            data = new Data("/u/mammobayes/DATA/matched outcome/jesse's/notreport/birads/test", ".data", dataSets, test, "/u/mammobayes/DATA/matched outcome/jesse's/notreport/birads/bc-new.names", ",", true, true);
        }
        else if (useAggs == 10){
            System.out.println("Creating initial structure only radiologist + aggregate features with matched data");

            data = new Data("/u/mammobayes/DATA/matched outcome/jesse's/notreport/sayu/aggregate/", "trainAgg" + test, "tuneAgg" + test, "testAgg" + test, ".data",".key", "fold" + test + ".names", ",", true, true);


        }
        else if (useAggs == 11){
            //data = null;
            //Flag - this should really 
            System.out.println("Creating initial structure with radiologist + aggregate features on matched data (includes birads)");
            data = new Data("/u/mammobayes/DATA/matched outcome/jesse's/notreport/birads/sayu/aggregate/", "trainAgg" + test, "tuneAgg" + test, "testAgg" + test, ".data",".key", "fold" + test + ".names", ",", true, true);
            //data = new Data("/u/jdavis/Yap/ILP/AlephUW/dev/data/old/level2/tmp/agg/", "trainAgg" + test, "tuneAgg" + test, "testAgg" + test, ".data",".key", "fold" + test + ".names", ",", true, true);
        }
        else if (useAggs == 12){
            System.out.println("Creating initial structure only radiologist + aggregate features with matched data");

            data = new Data("/u/mammobayes/DATA/matched outcome/jesse's/none/sayu/aggregate/", "trainAgg" + test, "tuneAgg" + test, "testAgg" + test, ".data",".key", "fold" + test + ".names", ",", true, true);


        }
        else if (useAggs == 13){
            //data = null;
            //Flag - this should really 
            System.out.println("Creating initial structure with radiologist + aggregate features on matched data (includes birads)");
            data = new Data("/u/mammobayes/DATA/matched outcome/jesse's/none/birads/sayu/aggregate/", "trainAgg" + test, "tuneAgg" + test, "testAgg" + test, ".data",".key", "fold" + test + ".names", ",", true, true);
            //data = new Data("/u/jdavis/Yap/ILP/AlephUW/dev/data/old/level2/tmp/agg/", "trainAgg" + test, "tuneAgg" + test, "testAgg" + test, ".data",".key", "fold" + test + ".names", ",", true, true);
        }
        else{
            data = null;
        }
        return data;
    }
    /** 
       this code goes through the graph and finds all of the parents 
       of the current vertex being examined
       @param currNode = the node to find parents for
       @param graph - the current network
       @param inDegrees - indegrees for each node
       @return an array which holds the positions of the parents in the
       feature vector
     */
    public static int[] findParents(int currNode, int[][] graph){

        int inDegree = 0;
        for(int ix = 0; ix < graph.length;ix++){
            if (graph[currNode][ix] == Constants.IN_EDGE){
                inDegree++;
            }
        }
        //array to hold positions (in attribute array) of parents

        int[] parPos = new int[inDegree];
        int kx = 0;
        boolean one = false; //a cut out after finding a parent
        //go through and find which attributes are parents
        for(int jx = 0; jx < parPos.length; jx++){
            for(; kx < graph.length && !one; kx++){
                if (graph[currNode][kx] == Constants.IN_EDGE){
                    parPos[jx] = kx;
                    one = true;
                }//end of if (edges[currNode][kx] == IN_EDGE)
            }//end of kx for loop
            one = false;
        }//end of j for loop

        return parPos;


    }//end of method findParents

    /**
       @param graph
       @param outDegrees
       @param 
       @return true if the graph is acyclic
     */
    public static boolean acyclic(int[][] graph, int[] outDegrees, 
            int[] inDegrees, int startNode, 
            boolean directed){
        if (startNode >= 0){
            if (inDegrees[startNode] == 0){
                return true;
            }
        }//
        int[][] tempGraph = (int[][])graph.clone();
        int vertex = tempGraph.length;
        boolean failed = false;
        boolean[] valid = new boolean[tempGraph.length];
        int[] tempOutDegrees = (int[]) outDegrees.clone();

        for (int ix = 0; ix < valid.length; ix++){
            valid[ix] = true;
        }
        while ((vertex > 0) && (!failed)){
            int pos = 0;
            boolean hasLeaf = false;
            for(; pos < tempOutDegrees.length && !hasLeaf; pos++){

                if (directed){
                    if ((tempOutDegrees[pos] == 0) && (valid[pos])){
                        hasLeaf = true;
                        vertex--;
                        valid[pos] = false;
                    }
                }
                else{
                    if ((tempOutDegrees[pos] <= 1 ) && (valid[pos])){
                        hasLeaf = true;
                        vertex--;
                        valid[pos] = false;
                    }
                }
            }//end of for loop

            if (!hasLeaf){
                //System.out.println("in here..why?");
                failed = true;
                continue;
            }
            /*
	      remove the vertex at pos from the graph and all the edges
	      entering the vertex
             */

            pos--;
            for(int i = 0; i <tempGraph.length && pos < tempGraph.length; i++){

                if (tempGraph[pos][i] == Constants.IN_EDGE){
                    tempOutDegrees[i ] = tempOutDegrees[i ] - 1;
                }
            }//end for
        }//end while

        return !failed;
    }//end method acyclic

}
