%
% List of scoring functions,
% add yours here!
%
% or add them elsewhere, see if I care :)
%

:- module(cmp_score,[better_abs_score_variance_bias/4,
		     better_abs_score_variance_bias_with_threshold/4,
		     better_rel_score_variance_bias/4,
		     list_increased_score/2]).
		     
:- use_module('../common/utils',[geometric_average/2,average_and_stdev/3]).

:- dynamic min_improvement_threshold/1.

min_improvement_threshold(1.1).


%
% impose constraint on bias (average improvement) and variance (std dev).
%
better_abs_score_variance_bias(ScoreF, Score, MinVar, MaxBias) :-
	subtract_all(ScoreF, Score, Diffs),
	average_and_stdev(Diffs, Average, StDev),
	Average >= MinVar,
	StDev =< MaxBias*Average.

better_abs_score_variance_bias_with_threshold(ScoreF, Score, _MinVar, _MaxBias) :-
	min_improvement_threshold(MinOffset),
	subtract_all(ScoreF, Score, Diffs),
	average_and_stdev(Diffs, Average, StDev),
	Average > 0.0,
	Average >= StDev*MinOffset.

better_rel_score_variance_bias(ScoreF, Score, _MinVar, _MaxStdev) :-
	rel_all(ScoreF, Score, Diffs),
%	average_and_stdev(Diffs, Average, StDev),
%	geometric_average(Diffs, GAverage),
	negs(Diffs, 0, Negs),
%	format(user_error,'~w/~w & ~w~n',[MinVar,Average,Negs]),
%	Average >= MinVar,
	Negs =< 2.

list_increased_score(ScoreF, Score) :-
	number(ScoreF), !,
	ScoreF > Score.
list_increased_score(ScoreF, Score) :-
	sumlist(ScoreF, LF),
	sumlist(Score, L),
	LF > L.

subtract_all([], [], []).
subtract_all(S0.ScoreF, S1.Score, D.Diffs) :-
	D is S0-S1,
	subtract_all(ScoreF, Score, Diffs).

rel_all([], [], []).
rel_all(S0.ScoreF, S1.Score, D.Diffs) :-
	D is (S0-S1)/S1,
	rel_all(ScoreF, Score, Diffs).


negs([], Negs, Negs).
negs(D.Diffs, Negs0, Negs) :-
	(
	  D =< 0
	->
	  NegsI is Negs0+1
	;
	  NegsI = Negs0
	),
	negs(Diffs, NegsI, Negs).
