
:- module(jpl_interface,[myjpl_call/4,
			 myjpl_new/3,
			 myjpl_list_to_array/3,
			 zero_array/2]).

%
% files that scorer need before we do anything with java.
%
init_jvm :-
	prolog_load_context(directory,Dir),
	atom_concat(Dir,'/../sayu.jar:/../weka.jar',JPath),
	add_classes(JPath).

add_classes(JPath) :-
	unix(environ('CLASSPATH',CP)), !,
	atom_concat([JPath,':',CP],NEWCLASSPATH),
	unix(putenv('CLASSPATH',NEWCLASSPATH)).
add_classes(JPath) :-
	unix(putenv('CLASSPATH',JPath)).

% first do this, then load java
:- init_jvm.

%
% now, consult java
%
:- reexport(library(jpl),
	    [jpl_list_to_array/2,
	     jpl_list_to_array/3,
	     jpl_array_to_list/2]).

myjpl_list_to_array( Vals, Type, Obj) :-
	DebugLevel = 3,
	length(Vals, Len),
	jpl_list_to_array( Vals, Type, Obj),
	debug_format(DebugLevel,"NEW ~w[~w] -> ~w~n",[Type,Len,Obj]).

:- use_module('../common/debug_write',[debug_format/3]).

:-
	(
	  predicate_property(jpl_list_to_array(_,_,_),static)
	->
	  true
	;
	  assert_static((jpl_list_to_array(List, _, Obj) :-
			jpl_list_to_array(List, Obj)))
	).

myjpl_call(A,B,C,D) :-
	DebugLevel = 3,
	debug_format(DebugLevel,'JPL ~w:~q(~w) ',[A,B,C]),
	catch(jpl_call(A,B,C,D),Error,my_jplcall_error(A,B,C,D,Error)),
	debug_format(DebugLevel,'-> ~w~n',[D]).

myjpl_new(A,B,C) :-
	DebugLevel = 3,
	debug_format(DebugLevel,'JPL NEW ~w:~q() ',[A,B]),
	catch(jpl_new(A,B,C),Error,my_jplnew_error(A,B,C,Error)),
	debug_format(DebugLevel,'-> ~w~n',[C]).

my_jplcall_error(A,B,C,D,Error) :-
	format(user_error,'myjpl_call[~w,~w,~w,~w]. at ~w~n',[A,B,C,D,Error]).

my_jplnew_error(A,B,C,Error) :-
	format(user_error,'myjpl_new[~w,~w,~w]. at ~w~n',[A,B,C,Error]).


zero_array(Tot, Array) :-
 	zero_list(0, Tot, List),
	jpl_list_to_array(List, byte, Array).

zero_list(Tot, Tot, []) :- !.
zero_list(I0, Tot, 0.List) :-
	I is I0+1,
 	zero_list(I, Tot, List).

