%
% sayu feature evaluator
%
% all communication between JavaSayu and the rest of the world
% should be through this file.
%
% designed to support the old, class-based, interface.
%
%

:- module(sayu_attributes, [sayu_remove_feature/2,
			    sayu_replace_feature/6,
			    sayu_new_feature/5,
			    sayu_current_test_score/1,
			    sayu_current_train_score/1,
			    sayu_commit_update/1,
			    sayu_mark_committed/1,
			    sayu_get_cuts/1,
			    sayu_covers_example/4,
			    sayu_new_search/1,
			    sayu_scoring_data/4,
			    sayu_init_scoring/1,
			    sayu_init/0,
			    sayu_exit/0,
			    sayu_current_score/1,
			    sayu_vals_to_scorable/2,
			    sayu_count_cover/2,
			    sayu_disjoin/3,
			    sayu_renew_object/0,
			    sayu_vals_to_scorable/3,
			    sayu_get_probs/2]).

:- use_module(scorer, [cur_likelihood/1]).

:- use_module('../common/utils',[randomize_list/2]).

:- use_module('../common/set',[vsc_set/1]).

:- use_module('../common/debug_write',[debug_format/3]).

%
% fold loading 
% 
:- use_module(load_folds,[get_folds/6]).

%
% reset computational object
% 
:- use_module(retrain,[retrain_model/0]).

:- use_module(library(lists),[sumlist/2,
			      append/2,
			      member/2]).

:- use_module(jpl_interface).

:- dynamic sayu_java_arrays/6.

:- dynamic sayu_java_sizes/3.

:- dynamic sayu_object/1.

:- dynamic sayu_class/1.

% make object oriented
sayu_class(0, 'sayu.IncNB'). % IncNB
sayu_class(1, 'sayu.IncTan'). % IncTan

% scoring method:
% 0: weighted accuracy
% 1: Precision-Recall
% 2: AUC

:- dynamic sayu_scoring_method/1.

sayu_scoring_method(1).

% Classifier type
% currently only jtree
:- dynamic sayu_classifier_type/1.

sayu_classifier_type(4).

:- dynamic sayu_psuedo_count/1.

sayu_psuedo_count(1).

:- dynamic sayu_n_of_evaluated_atts/1.

sayu_n_of_evaluated_atts(0).

:- dynamic sayu_use_aggs/1.

sayu_use_aggs(0).

:- dynamic start_of_time/2.

start_of_time(0,0).	

%
% tell WEKA that you don't want 
% now permanently removes something out of network
sayu_remove_feature(Id, ScoreR) :-
	sayu_object(Object),
   	sayu_scoring_method(Method),
	debug_format(1,'-DJ- REPLACE ATTRIBUTE ~w~n',[Id]),
	myjpl_call(Object, removeAttribute, [Id, Method], ScoreR).

% doesn't exist any longer.
sayu_replace_feature(Id, TrPos-TuPos, TrNeg-TuNeg, TePos, TeNeg, Score) :-
   	sayu_object(Object),
   	sayu_scoring_method(Method),
	debug_format(2,'-DJ- REPLACE ATTRIBUTE ~w~n',[Id]),
	myjpl_call(Object, replaceAttribute, [TrPos, TrNeg, TuPos, TuNeg, TePos, TeNeg, Id, Method, 2], Score).


sayu_current_test_score(Score) :-
	sayu_object(Object), 
	sayu_scoring_method(ScoringMethod),
	myjpl_call(Object, getScore, [2,ScoringMethod], Score).

sayu_current_train_score(Score) :-
	sayu_object(Object), 
	sayu_scoring_method(ScoringMethod),
	myjpl_call(Object, getScore, [0,ScoringMethod], Score).

sayu_mark_committed(Score) :-
	sayu_object(Object),
	myjpl_call(Object, commit, [], _),
	retractall(cur_likelihood(_)),
	assert(cur_likelihood(Score)).

sayu_commit_update(Score) :-
	retractall(cur_likelihood(_)),
	assert(cur_likelihood(Score)),
	sayu_object(Object),
	myjpl_call(Object, commit, [], _),
%	sayu_scoring_method(ScoringMethod),
	sayu_dump_plist.

sayu_get_cuts(Thresholds) :-
	sayu_object(Object),
	myjpl_call(Object, getCuts, [], ThresholdsObj),
	jpl_array_to_list(ThresholdsObj, Thresholds).

sayu_new_feature(_,_,_,_, _) :-
	retract(sayu_n_of_evaluated_atts(Atts)),
	Atts1 is Atts+1,
	assert(sayu_n_of_evaluated_atts(Atts1)),
	fail.
sayu_new_feature(TrPos-TuPos,TrNeg-TuNeg,TePos,TeNeg, Likelihood) :-
	sayu_scoring_method(Method),
	sayu_object(Object),
	myjpl_call(Object, addAttribute2, [TrPos,TrNeg,TuPos,TuNeg,TePos,TeNeg,Method,2], LikelihoodObj), jpl_array_to_list(LikelihoodObj,Likelihood).

sayu_covers_example(TrainObj-TuneObj, PosNeg, _, Id) :- !,
	sayu_java_sizes(train, PosNeg, TrainPosSize ),
	(
	  Id > TrainPosSize
	->
	  Indx is Id-(TrainPosSize+1),
	  jpl:jni_SetByteArrayElement(TuneObj, Indx, 1)
	;
	  Indx is Id-1,
	  jpl:jni_SetByteArrayElement(TrainObj, Indx, 1)
	).
sayu_covers_example(Obj, _PosNeg, _Type, Id) :-
	Indx is Id-1,
	jpl:jni_SetByteArrayElement(Obj, Indx, 1).

sayu_new_search(Num) :-
	sayu_object(Object), 
	myjpl_call(Object,newHead,[Num],_).
	
sayu_scoring_data(TrP-TuP,TrN-TuN,TeP,TeN) :-
	sayu_java_arrays(TrP,TrN,TuP,TuN,TeP,TeN).

sayu_init_scoring(Obj1-Obj2) :- !,
	myjpl_call( 'sayu.MyArrays', reset, [Obj1], _),
	myjpl_call( 'sayu.MyArrays', reset, [Obj2], _).
sayu_init_scoring(Obj) :-
	myjpl_call( 'sayu.MyArrays', reset, [Obj], _).

sayu_current_score(Score) :-
	cur_likelihood(Score).

sayu_disjoin(Score, Rule1, Rule2) :-
	sayu_object(Object),
	sayu_scoring_method(ScoringMethod),	
	%debug_format(1, 'calling disjoin~n', []),
	myjpl_call(Object, disjoin, [ScoringMethod], Obj),
	%ThresholdsObj),
	jpl_array_to_list(Obj, [Score, Rule1, Rule2| _]).

sayu_count_cover(BMapTr-BMapTu, Cover) :- !,
	myjpl_call( 'sayu.MyArrays', sum, [BMapTr], SumR),
	myjpl_call( 'sayu.MyArrays', sum, [BMapTu], SumU),
	Cover is SumR+SumU.
sayu_count_cover(BMap, Cover) :-
	myjpl_call( 'sayu.MyArrays', sum, [BMap], Cover).

sayu_vals_to_scorable(Vals, Type, Obj) :-
	jpl_list_to_array( Vals, Type, Obj).

sayu_get_probs(DataSet, AllProbs) :-
	sayu_object(Object),
	myjpl_call(Object, getProbs, [DataSet], Probs),
	jpl_array_to_list(Probs, AllProbs).

sayu_init :-
	sayu_init_jvm,
	retractall(start_of_time(_,_)),
	statistics(cputime,[T0,_]),
	statistics(walltime,[WT0,_]),
	assert(start_of_time(T0,WT0)),
	sayu_do_init,
	retractall(sayu_n_of_evaluated_atts(_)),
	assert(sayu_n_of_evaluated_atts(0)).

sayu_init_jvm :-
	predicate_property(jpl:myjpl_call(_,_,_,_),_), !.
sayu_init_jvm :-
	sayu_add_sayu_classes.

:- prolog_load_context(directory,Dir),
	atom_concat([Dir,'/../sayu.jar:',Dir,'/../weka.jar'],JPath),
	assert_static(sayu_location(JPath)).

% for experimenting with hierarchy learning
sayu_add_sayu_classes :-
	unix(environ('CLASSPATH',CP)), !,
	sayu_location(JPath),
	atom_concat([JPath,':',CP],NEWCLASSPATH),
	unix(putenv('CLASSPATH',NEWCLASSPATH)).
sayu_add_sayu_classes :- 
	sayu_location(JPath),
	unix(putenv('CLASSPATH',JPath)).


sayu_do_init :-
%	myjpl_call('weka.NBayes', init, [PSize,NSize], Likelihood),
	retractall(cur_likelihood(_)),
	retractall(sayu_java_arrays(_,_,_,_,_,_)),
	retractall(sayu_java_sizes(_,_,_)),
	get_folds(PosTrainFolds, NegTrainFolds, PosTuneFolds, NegTuneFolds, [TestPos], [TestNegs]),
	sumlist(PosTrainFolds, TrainPos),
	sumlist(PosTuneFolds, TunePos),
	sumlist(NegTrainFolds, TrainNeg),
	sumlist(NegTuneFolds, TuneNeg),
	retractall(sayu_obj_settings(_, _, _, _)),
	assert(sayu_obj_settings(TrainPos, TrainNeg, TunePos, TuneNeg, TestPos, TestNegs)),
	sayu_scoring_method(ScoringMethod),
	sayu_classifier_type(ClassifierType),
	( true
	->
	  user:sayu_strata(NTrain0f,NTune0f,NTest0f,NTrain0n,NTune0n,NTest0n),
	  myjpl_new('sayu.DifferentialIncBN',[TrainPos, TrainNeg, TunePos, TuneNeg, TestPos, TestNegs, NTrain0f, NTune0f, NTest0f, NTrain0n, NTune0n, NTest0n, ClassifierType], SayuObj)
	;
	  sayu_class(ClassifierType, SayuClass)
	->
	  myjpl_new(SayuClass,[TrainPos, TrainNeg, TunePos, TuneNeg, TestPos, TestNegs], SayuObj)
	;
	  throw(unsupported_sayu_classifier(ClassifierType))
	),
	assert(sayu_object(SayuObj)),
	sayu_set_pr_area_threshold,
	myjpl_call(SayuObj, getScore, [2,ScoringMethod], InitialTestScore),
	myjpl_call(SayuObj, getScore, [1,ScoringMethod], InitialTuneScore),
	myjpl_call(SayuObj, getScore, [0,ScoringMethod], InitialTrainScore),
	format(user_error,'Initial Train/Tune/Test Score: ~p/~p/~p~n',[InitialTrainScore,InitialTuneScore,InitialTestScore]),
	zero_array(TrainPos, TrainPosArray),
	zero_array(TrainNeg, TrainNegArray),
	zero_array(TunePos, TunePosArray),
	zero_array(TuneNeg, TuneNegArray),
	zero_array(TestPos, TestPosArray),
	zero_array(TestNegs, TestNegArray),
	assert(sayu_java_arrays(TrainPosArray,TrainNegArray,TunePosArray,TuneNegArray,TestPosArray,TestNegArray)),
	assert(sayu_java_sizes(train, pos, TrainPos )),
	assert(sayu_java_sizes(train, neg, TrainNeg )),
	assert(sayu_java_sizes(tune, pos ,TunePos)),
	assert(sayu_java_sizes(tune, neg, TuneNeg)),
	assert(sayu_java_sizes(test, pos, TestPos)),
	assert(sayu_java_sizes(test, neg, TestNegs)),
	assert(cur_likelihood([InitialTrainScore,InitialTuneScore])),
	assert(user:sayu_old_best([InitialTrainScore,InitialTuneScore])).

sayu_exit :-
	flush_output,
	format(user_error,'% clauses in theory:~n',[]),
	fail.
sayu_exit :-
	user:sayu_output_theory(Oth),
	open(Oth, write, W),
	output_theory(W),
	close(W),
	fail.
sayu_exit :-
	recorded(theory_clause, Cl, _),
	portray_clause(user_error,Cl),
	fail.
sayu_exit :-
	sayu_object(Obj),
	sayu_scoring_method(ScoringMethod),
	myjpl_call(Obj, getScore, [2,ScoringMethod], TestLikelihood),
	format(user_error,'[ Final Test Score ~4f ]~n',[TestLikelihood]),
	fail.
sayu_exit :-
	start_of_time(T0, WT0),
	statistics(cputime,[_,T]),
	DT is T-T0,
	format(user_error,'[ CPU Time ~d msec. ]~n',[DT]),
	statistics(walltime,[_,WT]),
	WDT is WT-WT0,
	format(user_error,'[ Wall Time ~d msec. ]~n',[WDT]),
	fail.
sayu_exit :-
	% show current performance
	sayu_class(Obj),
	sayu_scoring_method(ScoringMethod),
	myjpl_call(Obj, printCurve, [2,ScoringMethod], _),
	fail.
sayu_exit :-
	% show current performance
%	sayu_class(Class),
%	sayu_scoring_method(ScoringMethod),
%	myjpl_call(Class, printCurve, [1,ScoringMethod], _),
%	myjpl_call(Class, printCurve, [2,ScoringMethod], _),
	fail.
sayu_exit :-
%	sayu_dump_plist,
	halt.

sayu_dump_plist :-
	( user:sayu_dump_file(F) -> open(F,write,S) ; S = user_error ),
	sayu_object(Obj),
	myjpl_call(Obj, getProbs, [2], Probs),
	once(current_predicate(_,test_pos:P)),
	findall(P,test_pos:P,LP),
	%writeln(LP),
	findall(P,test_neg:P,LN),
	jpl_array_to_list(Probs, AllProbs),
%	length(AllProbs,L0),
%	length(LP,LLP),
%	length(LN,LLN),
%	format(user_error,'probs: ~d ~d ~d ~n',[LLP,LLN,L0]),
	sayu_dump_probs(LP,pos,AllProbs,S,Rest),
	sayu_dump_probs(LN,neg,Rest,S,[]),
	close(S).

sayu_dump_probs([],_,Probs,_,Probs).
sayu_dump_probs([Ex|LP],Type,[P|AllProbs],S,Rest) :-
	format(S,'prob(~w,~q,~w).~n',[Type,Ex,P]),
	sayu_dump_probs(LP,Type,AllProbs,S,Rest).


sayu_set_pr_area_threshold :-
	user:sayu_min_recall_in_area(Threshold),
	sayu_object(Object),
	myjpl_call(Object, 'setAUCPRMinRecall', [Threshold], _).

sayu_renew_object :-
	throw('renew_object unsupported with this interface').

output_theory(W) :-
	recorded(theory_clause, Cl, _),
	portray_clause(W, Cl),
	fail.
output_theory(_).


