%
% sayu feature evaluator
%
% all communication with the rest of sayu should be through this file.
%

:- module(score_attributes, [remove_feature/2,
			     replace_feature/6,
			     new_feature/5,
			     commit_update/1,
			     mark_committed/1,
			     renew_object/0,
			     get_cuts/1,
			     covers_example/4,
			     init_scoring/1,
			     new_search/1,
			     scoring_data/4,
			     better_score/2,
			     better_hierarchy_score/2,
			     better_entry_score/2,
			     increased_score/2,
			     scorer_init/0,
			     scorer_exit/0,
	 		     scorer_new_seed/0,
			     current_score/1,
			     current_train_score/1,
			     current_test_score/1,
			     count_cover/2,
			     vals_to_scorable/3,
			     get_probs/2,
			     disjoin/3,
			     cur_likelihood/1,
			     scorer_has_tuneset/0]).

:- use_module(cvsayu, [cvsayu_remove_feature/2,
		       cvsayu_replace_feature/6,
		       cvsayu_new_feature/5,
		       cvsayu_commit_update/1,
		       cvsayu_mark_committed/1,
		       cvsayu_renew_object/0,
		       cvsayu_get_cuts/1,
		       cvsayu_covers_example/4,
		       cvsayu_new_search/1,
		       cvsayu_scoring_data/4,
		       cvsayu_init_scoring/1,
		       cvsayu_init/0,
		       cvsayu_exit/0,
		       cvsayu_current_score/1,
		       cvsayu_current_train_score/1,
		       cvsayu_current_test_score/1,
		       cvsayu_count_cover/2,
		       cvsayu_disjoin/3,
		       cvsayu_vals_to_scorable/3,
		       cvsayu_get_probs/2
		      ]).

:- use_module(weka, [weka_remove_feature/2,
		       weka_replace_feature/6,
		       weka_new_feature/5,
		       weka_commit_update/1,
		       weka_mark_committed/1,
		       weka_renew_object/0,
		       weka_get_cuts/1,
		       weka_covers_example/4,
		       weka_new_search/1,
		       weka_scoring_data/4,
		       weka_init_scoring/1,
		       weka_init/0,
		       weka_exit/0,
		       weka_current_score/1,
		       weka_current_train_score/1,
		       weka_current_test_score/1,
		       weka_count_cover/2,
		       weka_disjoin/3,
		       weka_vals_to_scorable/3,
		       weka_get_probs/2
		      ]).

:- use_module(sayu, [sayu_remove_feature/2,
		       sayu_replace_feature/6,
		       sayu_new_feature/5,
		       sayu_commit_update/1,
		       sayu_mark_committed/1,
		       sayu_renew_object/0,
		       sayu_get_cuts/1,
		       sayu_covers_example/4,
		       sayu_new_search/1,
		       sayu_scoring_data/4,
		       sayu_init_scoring/1,
		       sayu_init/0,
		       sayu_exit/0,
		       sayu_current_score/1,
		       sayu_current_train_score/1,
		       sayu_current_test_score/1,
		       sayu_count_cover/2,
		       sayu_disjoin/3,
		       sayu_vals_to_scorable/3,
		       sayu_get_probs/2
		      ]).

:- use_module(cmp_scores, [better_abs_score_variance_bias/4,
			   better_abs_score_variance_bias_with_threshold/4,
			   better_rel_score_variance_bias/4,
			   list_increased_score/2]).

:- dynamic cur_likelihood/1.

:- dynamic cur_search/1.

:- dynamic(scorer/1).

scorer(sayu).

:- dynamic(target/1).

target(1.02).

:- dynamic(min_bias/1).

min_bias(0.02).

:- dynamic(max_stdev/1).

max_stdev(1.0).

:- dynamic(cmp_scores_howto/1).

cmp_scores_howto(rel_variance_bias).

remove_feature(Id, NewScore) :-
	scorer(Scorer),
	(
	  Scorer == cvsayu
	->
	  cvsayu_remove_feature(Id, NewScore)
	;
	  Scorer == sayu
	->
	  sayu_remove_feature(Id, NewScore)
	;
	  Scorer == weka
	->
	  weka_remove_feature(Id, NewScore)
	;
	 throw(error(unknown_scorer(Scorer), remove_feature(Id,NewScore)))
	).
	  
replace_feature(Id,TrPos,TrNeg,TePos,TeNeg, Scores) :-
	scorer(Scorer),
	(
	  Scorer == cvsayu
	->
	  cvsayu_replace_feature(Id,TrPos,TrNeg,TePos,TeNeg, Scores)
	;
	  Scorer == sayu
	->
	  sayu_replace_feature(Id,TrPos,TrNeg,TePos,TeNeg, Scores)
	;
	  Scorer == weka
	->
	  weka_replace_feature(Id,TrPos,TrNeg,TePos,TeNeg, Scores)
	;
	 throw(error(unknown_scorer(Scorer), replace_feature(Id,TrPos,TrNeg,TePos,TeNeg, Scores)))
	).
	  
new_feature(TrPos,TrNeg,TePos,TeNeg, Scores) :-
	scorer(Scorer),
	(
	  Scorer == cvsayu
	->
	  cvsayu_new_feature(TrPos,TrNeg,TePos,TeNeg, Scores)
	;
	  Scorer == sayu
	->
	  sayu_new_feature(TrPos,TrNeg,TePos,TeNeg, Scores)
	;
	  Scorer == weka
	->
	  weka_new_feature(TrPos,TrNeg,TePos,TeNeg, Scores)
	;
	 throw(error(unknown_scorer(Scorer), new_feature(TrPos,TrNeg,TePos,TeNeg, Scores)))
	).

current_train_score(Score) :-
	scorer(Scorer),
	(
	  Scorer == cvsayu
	->
	  cvsayu_current_train_score(Score)
	;
	  Scorer == sayu
	->
	  sayu_current_train_score(Score)
	;
	  Scorer == weka
	->
	  weka_current_train_score(Score)
	;
	 throw(error(unknown_scorer(Scorer), current_test_score(Score)))
	).
	  
current_test_score(Score) :-
	scorer(Scorer),
	(
	  Scorer == cvsayu
	->
	  cvsayu_current_test_score(Score)
	;
	  Scorer == sayu
	->
	  sayu_current_test_score(Score)
	;
	  Scorer == weka
	->
	  weka_current_test_score(Score)
	;
	 throw(error(unknown_scorer(Scorer), current_test_score(Score)))
	).
	

:- dynamic score/1, sq/3.

new_variance :-
	score([_,_]),
	retractall(sq(_,_,_)),
	findall(X, score(X._), ST),
	findall(X, retract(score(_.X)), S),
	S = [_,_|_], !,
	avg(ST, 0, 0.0, AvT),
	variance(ST, -1, 0.0, AvT, VarT),
	SqDevT is sqrt(VarT),
	assert(sq(train,AvT,SqDevT)),
margin(train,AvT,2,MarginT), writeln((avt:AvT:SqDevT->MarginT)),
	avg(S, 0, 0.0, Av),
	variance(S, -1, 0.0, Av, Var),
	SqDev is sqrt(Var),
	assert(sq(tune,Av,SqDev)),
margin(tune,Av,1,Margin), writeln((av:Av:SqDev->Margin)).
new_variance :-
	score(A),
	number(A),
	retractall(sq(_,_)),
	findall(X, retract(score(X)), S),
	S = [_,_|_], !,
	avg(S, 0, 0.0, Av),
	variance(S, -1, 0.0, Av, Var),
	SqDev is sqrt(Var),
	assert(sq(tune,Av,SqDev)),
margin(tune,Av,1,Margin), writeln((av:Av:SqDev->Margin)).
new_variance.

avg([], N, Sum, Av) :-
	Av is Sum/N.
avg(E.S, N0, Sum0, Av) :-
	N is N0+1,
	Sum is Sum0+E,
	avg(S, N, Sum, Av).

variance([], N, Sum, _, Av) :-
	Av is Sum/N.
variance(E.S, N0, Sum0, Avg, Av) :-
	N is N0+1,
	Sum is Sum0+(E-Avg)*(E-Avg),
	variance(S, N, Sum, Avg, Av).

  
commit_update(Score) :-
	scorer(Scorer),
	(
	  Scorer == cvsayu
	->
	  cvsayu_commit_update(Score)
	;
	  Scorer == sayu
	->
	  sayu_commit_update(Score)
	;
	  Scorer == weka
	->
	  weka_commit_update(Score)
	;
	 throw(error(unknown_scorer(Scorer), commit_update(Score)))
	).
	  
mark_committed(Scores) :-
	scorer(Scorer),
	(
	  Scorer == cvsayu
	->
	  cvsayu_mark_committed(Scores)
	;
	  Scorer == sayu
	->
	  sayu_mark_committed(Scores)
	;
	  Scorer == weka
	->
	  weka_mark_committed(Scores)
	;
	 throw(error(unknown_scorer(Scorer), mark_commited(Scores)))
	).
	  
renew_object :-
	scorer(Scorer),
	(
	  Scorer == cvsayu
	->
	  cvsayu_renew_object
	;
	  Scorer == sayu
	->
	  sayu_renew_object
	;
	  Scorer == weka
	->
	  weka_renew_object
	;
	 throw(error(unknown_scorer(Scorer), renew_object))
	).
	  
get_cuts(Thresholds) :-
	scorer(Scorer),
	(
	  Scorer == cvsayu
	->
	  cvsayu_get_cuts(Thresholds)
	;
	  Scorer == sayu
	->
	  sayu_get_cuts(Thresholds)
	;
	  Scorer == weka
	->
	  weka_get_cuts(Thresholds)
	;
	 throw(error(unknown_scorer(Scorer), get_cuts(Thresholds)))
	).
	  
covers_example(Obj, PosNeg, Type, Id) :-
	scorer(Scorer),
	(
	  Scorer == cvsayu
	->
	  cvsayu_covers_example(Obj, PosNeg, Type, Id)
	;
	  Scorer == sayu
	->
	  sayu_covers_example(Obj, PosNeg, Type, Id)
	;
	  Scorer == weka
	->
	  weka_covers_example(Obj, PosNeg, Type, Id)
	;
	 throw(error(unknown_scorer(Scorer), covers_example(Obj, PosNeg, Type, Id)))
	).
	  
new_search(Num) :-
	scorer(Scorer),
	(
	  Scorer == cvsayu
	->
	  cvsayu_new_search(Num)
	;
	  Scorer == sayu
	->
	  sayu_new_search(Num)
	;
	  Scorer == weka
	->
	  weka_new_search(Num)
	;
	 throw(error(unknown_scorer(Scorer), new_search))
	).
	  
scoring_data(TrP, TrN, TeP, TeN) :-
	scorer(Scorer),
	(
	  Scorer == cvsayu
	->
	  cvsayu_scoring_data(TrP, TrN, TeP, TeN)
	;
	  Scorer == sayu
	->
	  sayu_scoring_data(TrP, TrN, TeP, TeN)
	;
	  Scorer == weka
	->
	  weka_scoring_data(TrP, TrN, TeP, TeN)
	;
	 throw(error(unknown_scorer(Scorer), scoring_data(TrP, TrN, TeP, TeN)))
	).

get_probs(DataSet, AllProbs) :-
	scorer(Scorer),
	(
	  Scorer == cvsayu
	->
	  cvsayu_get_probs(DataSet, AllProbs)
	;
	  Scorer == sayu
	->
	  sayu_get_probs(DataSet, AllProbs)
	;
	  Scorer == weka
	->
	  weka_get_probs(DataSet, AllProbs)
	;
	 throw(error(unknown_scorer(Scorer), get_probs(DataSet, AllProbs)))
	).

scorer_new_seed :-
	new_variance.

init_scoring(Obj) :-
	scorer(Scorer),
	(
	  Scorer == cvsayu
	->
	  cvsayu_init_scoring(Obj)
	;
	  Scorer == sayu
	->
	  sayu_init_scoring(Obj)
	;
	  Scorer == weka
	->
	  weka_init_scoring(Obj)
	;
	 throw(error(unknown_scorer(Scorer), init_scoring(Obj)))
	).

init_scoring_data(Type1, TypeA, Set, Sign, CountObj) :-
	scorer(Scorer),
	(
	  Scorer == cvsayu
	->
	  cvsayu_init_scoring_data(Type1, TypeA, Set, Sign, CountObj)
	;
	  Scorer == sayu
	->
	  sayu_init_scoring_data(Type1, TypeA, Set, Sign, CountObj)
	;
	  Scorer == weka
	->
	  weka_init_scoring_data(Type1, TypeA, Set, Sign, CountObj)
	;
	 throw(error(unknown_scorer(Scorer), init_scoring_data(Type1, TypeA, Set, Sign, CountObj)))
	).


init_scoring_data(Type1, Type2, TypeA, Set, Sign, CountObj) :-
	scorer(Scorer),
	(
	  Scorer == cvsayu
	->
	  cvsayu_init_scoring_data(Type1, Type2, TypeA, Set, Sign, CountObj)
	;
	  Scorer == sayu
	->
	  sayu_init_scoring_data(Type1, Type2, TypeA, Set, Sign, CountObj)
	;
	  Scorer == weka
	->
	  weka_init_scoring_data(Type1, Type2, TypeA, Set, Sign, CountObj)
	;
	 throw(error(unknown_scorer(Scorer), init_scoring_data(Type1, Type2, TypeA, Set, Sign, CountObj)))
	).

significance_threshold(2.0).

% compute the Margin using a Gumbel distribtion. Use Factor 2 to give less
% weight to train data.
margin(Set,OldScore, Factor, Margin) :-
	fail,
	sq(Set,Avg,Dev), !,
	Beta is (sqrt(6.0)/pi)*Dev,
	Nu0 is Avg+0.5772*Beta,
	Nu is max(Nu0, OldScore),
	significance_threshold(T),
	Margin is Nu+T*Beta/Factor.
margin(_Set,OldScore, Factor, N) :- N is OldScore * (1.0+0.02/Factor).

export_margin(Margin0, OldScore, _Beta, _Factor, Margin) :-
	OldScore < Margin0, !,
	Margin = Margin0.
% We have improved over old score (hierarchy). 
% Assume this is the new average
export_margin(Margin0, OldScore, _Beta, _Factor, Margin) :-
	Nu is OldScore+0.5772*_Beta,
	hie_significance_threshold(T),
	OldMargin is Nu+T*_Beta/_Factor,
	Margin is max(Margin0, OldMargin).

better_hierarchy_score([NewScoreT,NewScore], [OldScoreT,OldScore]) :-	
	target(Par),
	( vista_constants:demand_constants(true) -> Target is Par+(Par-1.02)*2 ;
	  Target is Par-(Par-1.0)/2
	),
	NewScore > OldScore*Target,	
	NewScoreT > OldScoreT*Target,	
	Max is max(NewScore,NewScoreT),
	Min is min(NewScore,NewScoreT),
	Max/Min < 1.5.

better_entry_score([NewScoreT,NewScore], [OldScoreT,OldScore]) :-
	target(Par),
	Target is 1.0+(Par-1.0)/3.0,
	NewScore > OldScore*Target,	
	NewScoreT > OldScoreT*Target,	
	Max is max(NewScore,NewScoreT),
	Min is min(NewScore,NewScoreT),
	Max/Min < 1.5.

better_score(NewScore, OldScore) :-
	number(NewScore), !,	
	vista:step(N),
	(
	  sayu_attributes:sayu_scoring_method(0)
	->
	    (0.5-NewScore) < (1-0.1/(N+1))*(0.5-OldScore)
	;
%	    min_bias(Bias),
%	    NewScore > OldScore, 	
	    	    NewScore > OldScore*(1.21-0.2/(1+exp(-N/2)))
%	    	    NewScore > OldScore*(1.31-0.3/(1+exp(-N/2)))
%	margin(tune,OldScore,1, M), NewScore> M	
	).
better_score([NewScoreT,NewScore], [OldScoreT,OldScore]) :- !,
	%writeln(([NewScoreT,NewScore], [OldScoreT,OldScore])),
	%sayu_attributes:sayu_scoring_method(6), !,
	user:sayl_threshold(Par), % = 1.05, 1.01, 1.1,
	NewScore> Par*OldScore,	
	NewScoreT> Par*OldScoreT.
better_score([NewScoreT,NewScore], [OldScoreT,OldScore]) :- !,
	target(Par),
	NewScore> Par*OldScore,	
	NewScoreT> Par*OldScoreT,	
	Max is max(NewScore,NewScoreT),
	Min is min(NewScore,NewScoreT),
	Max/Min < 1.5.
better_score(NewScore, OldScore) :-
	min_bias(Bias),
	max_stdev(StDev),
	cmp_scores_howto(CmpScorer),
	(
	  CmpScorer == abs_variance_bias
	->
	  better_abs_score_variance_bias(NewScore, OldScore, Bias, StDev)
	;
	  CmpScorer == abs_variance_bias_with_threshold
	->
	  better_abs_score_variance_bias_with_threshold(NewScore, OldScore, Bias, StDev)
	;
	  CmpScorer == rel_variance_bias
	->
	  better_rel_score_variance_bias(NewScore, OldScore, Bias, StDev)
	).

current_score(Score) :-
	scorer(Scorer),
	(
	  Scorer == cvsayu
	->
	  cvsayu_current_score(Score)
	;
	  Scorer == sayu
	->
	  sayu_current_score(Score)
	;
	  Scorer == weka
	->
	  weka_current_score(Score)
	;
	 throw(error(unknown_scorer(Scorer), current_score(Score)))
	).

increased_score(NewScore, OldScore) :-
	number(NewScore), !,	
	NewScore > OldScore. %*1.005.
increased_score([NewScore,NewScoreT], [OldScore,OldScoreT]) :- !,
	NewScore > OldScore, %*1.005.
	NewScoreT > OldScoreT. %*1.005.
increased_score([NewScore|NewScoreT], [OldScore|OldScoreT]) :- !,
	NewScore > OldScore, %*1.005.
	NewScoreT > OldScoreT. %*1.005.
increased_score(NewScore, OldScore) :-
	list_increased_score(NewScore, OldScore).

scorer_init :-
	scorer(Scorer),
	(
	  Scorer == cvsayu
	->
	  cvsayu_init
	;
	  Scorer == sayu
	->
	  sayu_init
	;
	  Scorer == weka
	->
	  weka_init
	;
	 throw(error(unknown_scorer(Scorer), weka_init))
	).

scorer_exit :-
	scorer(Scorer),
	(
	  Scorer == cvsayu
	->
	  cvsayu_exit
	;
	  Scorer == sayu
	->
	  sayu_exit
	;
	  Scorer == weka
	->
	  weka_exit
	;
	 throw(error(unknown_scorer(Scorer), weka_exit))
	).


attributes_disjoin(Score, Rule1, Rule2) :-
	scorer(Scorer),
	(
	  Scorer == cvsayu
	->
	  cvsayu_disjoin(Score, Rule1, Rule2)
	;
	  Scorer == sayu
	->
	  sayu_disjoin(Score, Rule1, Rule2)
	;
	  Scorer == weka
	->
	  weka_disjoin(Score, Rule1, Rule2)
	;
	 throw(error(unknown_scorer(Scorer), attributes_disjoin(Score, Rule1, Rule2)))
	).


count_cover(BMap, Cover) :-
	scorer(Scorer),
	(
	  Scorer == cvsayu
	->
	  cvsayu_count_cover(BMap, Cover)
	;
	  Scorer == sayu
	->
	  sayu_count_cover(BMap, Cover)
	;
	  Scorer == weka
	->
	  weka_count_cover(BMap, Cover)
	;
	 throw(error(unknown_scorer(Scorer), count_cover(BMap, Cover)))
	).

vals_to_scorable(Vals, Type, Obj) :-
	scorer(Scorer),
	(
	  Scorer == cvsayu
	->
	  cvsayu_vals_to_scorable(Vals, Type, Obj)
	;
	  Scorer == sayu
	->
	  sayu_vals_to_scorable(Vals, Type, Obj)
	;
	  Scorer == weka
	->
	  weka_vals_to_scorable(Vals, Type, Obj)
	;
	 throw(error(unknown_scorer(Scorer), vals_to_scorable(Vals, Type, Obj)))
	).

scorer_has_tuneset :-
	scorer(Scorer),
	(
	  Scorer == sayu
	->
	  true
	;
	  Scorer == weka
	->
	  true
	).

