%:- yap_flag(agc_margin, 0).

%ignore above, will have modes and types as in aleph

%pulls off clause, scores it, builts all refinements of 
%clause
%:- style_check(-discontiguous).

:- yap_flag(unknown,error).

:- source.

:- module(vista,[main_timed/1,
	main/0,
	program/1,
	start_new_search/1,
	update_network/1,
	op(500,fy,#),
	op(500,fy,*),
	op(900,xfy,because)]).

%
% what vista does reexport to the world
%

:- reexport('vista/modes', [modeb/2,
	modeh/2,
	determination/2]).

:- reexport('vista/read_all', [read_all/1]).

:- reexport('vista/examples', [count/3]).

:- reexport('vista/aggregation', [linkage/3,
   linkage_compatible/1]).

:- reexport('vista/hierarchy', [build_hierarchy/1,reduce_hierarchy/0]).

:- reexport('common/set', [set/2,
	set/1,
	vsc_set/1,
	setting/2]).

% what vista uses.

:- use_module('vista/eval_node',[init_eval/0,
	evaluate_clause/4,
	final_eval_clause/6]).

:- use_module('vista/foil',[next_clause/1]).

:- use_module('vista/init',[view_init/0]).

:- use_module('vista/modes',[determination_arity/1,
	unassert_determination/1]).

:- use_module('vista/newpred',[assert_extensionally/3]).

:- use_module('vista/reduce',[init_reduce/3]).

:- use_module('vista/search_space',[init_search_space/0]).

:- use_module('vista/theory',[add_to_theory/2,
	list_theory/0]).

:- reexport('vista/bottom_clause',[constant_histogram/0]).

:- use_module('scorer/scorer', [commit_update/1,
	current_score/1,
	current_train_score/1,
	current_test_score/1,
	new_search/1,
	scorer_new_seed/0,
	scorer_exit/0]).

:- use_module('common/debug_write',[debug_format/3]).

:- use_module('common/random',[nextrandom/1]).

:- dynamic search_assert/1.

%:- style_check(single_var).

:- dynamic step/1.

step(0).

:- dynamic max_step/1.

max_step(-1).

:- dynamic itype_cur_num/1.

:- dynamic itype_2num/2.

:- dynamic use_hierarchy/1.

use_hierarchy(off).

:- dynamic use_disjoin/1.

use_disjoin(off).

:- dynamic rules_added/1.

rules_added(1).

:- dynamic split/1.

split(0.7).

:- dynamic invented_arity/1.

invented_arity(-1).

:- dynamic agg_nodes/1.

agg_nodes(200).

main_timed(T) :-
	view_init,
	alarm(T,scorer_exit,_),
	main,
%	test_disjoin_main,
	fail.
main_timed(_) :-
	scorer_exit.

main :-
	view_init,
	determination_arity(Arity),
	main(1,Arity).

main(Step,_Arity) :-
	max_step(Step), !.
main(Step,Arity) :-
%	(Nodes0 =< Max -> Nodes is Nodes0+Step ; Nodes=Nodes0),
	next_search(Step, Arity),
	% keep a running total of all processed clauses
	% used by scorer
	get_value(total_processed_clauses,Tot),
	get_value(processed_clauses,PTot),
	NTot is PTot+Tot,
	set_value(total_processed_clauses,NTot),
	% keep the current iteration
	Step1 is Step+1,
	retractall(step(_)),
	assert(step(Step1)),
	main(Step1,Arity).

next_search( Step, _) :-
	use_disjoin(disjoin),
	Step > 9,
	nextrandom(Y),
	Y > 0.9, !,
	current_score(Scores),
	%debug_format(1, 'trying to disjoin ~w~n', [Score]),
	init_eval,
	reduce_disjunction(Scores).
next_search(Steps, Arity) :-
	nextrandom(X),
	split(Split),
	X > Split, !,
	init_eval,
	setting(nodes,Nodes),
	agg_nodes(AggNodes),
	set(nodes, AggNodes),
	reduce_agg_clause(Steps, Arity),
	set(nodes, Nodes).
next_search(Steps, Arity) :-
	init_eval,
	reduce_clause(Steps, Arity).

% increase clause threshold

reduce_clause(Steps, Arity) :- 
	next_arity(Arity, N), !,
	reduce_clause(Steps, N, single).
reduce_clause(Steps, Arity) :- 
	reduce_clause(Steps, Arity).

reduce_agg_clause(Steps, Arity) :- 
	next_arity(Arity, N), !,
	N1 is N+1,
	reduce_clause(Steps, N1, aggs).
reduce_agg_clause(Steps, Arity) :- 
	reduce_agg_clause(Steps, Arity).

next_arity(_, N) :-
	invented_arity(N),
	N > 0, !.
next_arity(Arity, N) :-
	nextrandom(R),
	N is integer(R*Arity)+1.

update_network(Score) :-
	rules_added(Num),
	Num1 is Num +1,
	retractall(rules_added(_)),
	asserta(rules_added(Num1)),
	commit_update(Score),
	new_search(Num1).

reduce_clause(Steps, N, Aggs) :-
	length(L, N),
	functor(Type, t, N),
	% try until init_reduce finds something we like
	scorer_new_seed,
	init_reduce(L, N, Aggs),
	bfsearch(Type, Aggs),
	hierarchize(Steps).

hierarchize(Steps) :-
	use_hierarchy(on),
%	Steps > 20,
	Steps mod 3 =:= 0,
	!,
	reduce_hierarchy.
hierarchize(_).

bfsearch(Type, Aggs) :-
	start_new_search(Type),
	current_score(Scores),
	catch(refine_one(Type, Scores, Aggs),limit_exceeded,true),
	!,
	finish_search.
bfsearch(_, _) :-
	finish_search.

start_new_search(_) :-
	init_search_space.

cvt_itype_2num(Inps,Num) :-
	%debug_format(1, 'cvt~n', []),
	itype_2num(Inps,Num), !.
cvt_itype_2num(Inps,Num1) :-
	retract(itype_cur_num(Num)),
	%debug_format(1, 'retracting ~d~n', [Num]),
	Num1 is Num + 1,
	assert(itype_cur_num(Num1)),
	assert(itype_2num(Inps,Num1)).

itype_cur_num(0).

finish_search :-
	final_eval_clause(Clause,VarClause,_,_,Score,UseAggregates),
	!,
        update_network(Score),
        (search_assert(no_layer) ->
	 unassert_determination(VarClause),
	 NewVarClause = VarClause
	 ;
         assert_extensionally(VarClause, UseAggregates, NewVarClause)
	), 
%        assert_extensionally(VarClause, UseAggregates),
%	assert(VarClause),
	NewVarClause = (_PredHead :- _),
	current_train_score(TrainScore),
	current_test_score(TestScore),
        %debug_format(1,'PredHead ~w~n', [PredHead]),
	add_to_theory( NewVarClause, Clause),
	debug_format(1,'~n~n NEW VIEW (score ~w/~w/~w):~n    ~w~n~n',[TrainScore,Score,TestScore,Clause]),
	list_theory.
finish_search.


refine_one(Type, OldScore, Aggs) :-
	repeat,
	( next_clause(Clause) ->
	    evaluate_clause(Clause, Type, Aggs, OldScore)
	;
	   % hack to handle case where we don't have more clauses
	   !, fail
        ).

type_bin(1).
type_bin(2).
type_bin(3).

/* Remaining clauses are asserted by modeb */

