%
% This module implements the search space as a beam. It relies on YAP's 
%  non-backtrackable beam data-structure.
%
%
% Interface:
%  nit_search_space/0: final class to initialise beam.
%
%  add_to_beam/2: final class to store a new node in beam
%
%  get_from_beam/1: final class to retrieve a node from Beam.
%
%  close_beam/0: dispose of beam.
%
% Global Variables:
%
% beam_size: integer
%
%

:- module(vista_beam,[init_beam/0,
		      add_to_beam/2,
		      get_from_beam/1,
		      close_beam/0]).

:- use_module(library(nb)).

%
% make beam search the default search.
%

:- dynamic beam_size/1.

beam_size(4096).

init_beam :-
	close_beam,
	fail.
init_beam :-
	nb_beam(4096,Beam),
	nb_setval(beam, Beam).

close_beam :-
	catch(nb_getval(beam, Beam),_,fail),
	nb_beam_close(Beam),
	nb_delete(beam).

add_to_beam(K,V) :-
	nb_getval(beam, Beam),
	nb_beam_add(Beam, K, V).

get_from_beam(V) :-
	nb_getval(beam, Beam),
	nb_beam_del(Beam, _, V).
