%
% This module implements the search space as a beam in Prolog. It relies on
%  the internal data-base.
%
%
% Interface:
%  init_bestfirst/0: final class to initialise beam.
%
%  update_beam/2: final class to store a new node in beam
%
%  get_clause_beam/1: final class to retrieve a node from Beam.
%
% Global Variables:
%
% max_beam_size: integer
%
%

:- module(vista_bestfirst, [init_bestfirst/0,
			    update_beam/2,
			    get_clause_beam/1]).

:- use_module('../common/debug_write',[debug_format/3]).

:- use_module(library(lists),[append/3]).

:- use_module('../common/debug_write',[debug_format/3]).

:-dynamic(max_beam_size/1).

max_beam_size(2000).

:-dynamic(beam/1).
%beam([[0.1,(p(X):-r(X))], [0.2, (p(X):-q(X,Y),z(Y,Z))]]).

:-dynamic(beam_size/1).

beam_size(0).

init_bestfirst :-
	eraseall(beam),
	recorda(beam,[],_),
	retractall(beam_size(_)),
	assert(beam_size(0)).

remove_last([H],[], H).

remove_last([H|T], [H|List], Element) :-
	remove_last(T,List,Element).

get_clause_beam(Clause) :-
	recorded(beam,Beam,_),
	debug_format(2,'Beam ~w ~n', [Beam]),
	remove_last(Beam,Tail,[_,Clause]),
	debug_format(2,'Clause ~w ~n',[Clause]),
	eraseall(beam),
	recordz(beam,Tail,_).

update_beam(Score,Clause) :-
	beam(Beam),
%	write(Beam),
%	nl,
	new_beam(Beam,[Score,Clause],NewBeam),
%	write('New Beam '),
%	write(NewBeam),
%	nl,
%	beam_size(Size),
	length(NewBeam,Size),
	max_beam_size(MaxSize),
	eraseall(beam),
	(Size < MaxSize ->
		%Size1 is Size +1,	
		%retractall(beam_size(_)),
		%assert(beam_size(Size1)),
		recordz(beam,NewBeam,_)		
		;
		remove_first(NewBeam,NB),	
		recordz(beam,NB,_)
	),
	beam(B),
%	write('final beam '),
%	write(B),
%	debug_format(2,'New Beam ~w ~n',[B]),	
	update_beam_score(B).
	
update_beam_score([]) :-
	retractall(minScore(_)),
	assert(minScore(0.0)).
update_beam_score([[Score|_]|_]) :-
	retractall(minScore(_)),
	assert(minScore(Score)).

remove_first([],[]).
remove_first([_H|T], T).
	
new_beam([S1,Cl1].Beam,[Score,Clause],[S1,Cl1].NewBeam) :-
	Score < S1, !,
	new_beam(Beam,[Score,Clause],NewBeam).
new_beam(Beam,[Score,Clause],[Score,Clause].Beam).


beam(B) :-
	recorded(beam, B, _).
