%
% clause manipulation
%


:- module(vista_clause,[has_constant/1,
			replace_constant/5,
			has_end_constant/1,
			has_hierarchy/1,
			replace_end_constant/5]).

:- use_module(modes,[modeb/1]).

:- use_module(hierarchy,[ok_type/1]).

% The following version is marginally faster, but puts the disjuncts first
% contains_constant((H :- T), Const, Type, (H :- Disjunct,NewClause), Disjunct) :-
%    contains_constant_rec(T, Const, NewClause, Var, Type),
%    nonvar(Const),
%    get_new_disjunct(Name),
%    Disjunct =.. [Name, Var],
%    OrigFact =.. [Name, Const],
%    assert(OrigFact).

has_end_constant((_,T)) :- !,
	has_end_constant(T).
has_end_constant('!'(_,T)) :- !,
	has_end_constant(T).
has_end_constant(G) :-
	  has_constant(G).

%contains_constant_rec(T, Const, Changed, Var, Type) :-
%	rulelen(T,1),
%	contains_constant_rec2(T, Const, Changed, Var, Type).
has_constant((H,T)) :-
  	(
	  has_constant(H)
	;
	  has_constant(T)
	).
has_constant('!'(H,T)) :-
  	(
	  has_constant(H)
	;
	  has_constant(T)
	).
%contains_constant_rec(T, Const, Changed, Var, Type) :-
%	rulelen(T,1),
%	contains_constant_rec2(T, Const, Changed, Var, Type).

has_constant(H) :-
  	H =.. [Name|Body],
	cp_args(Body, Args),
  	Mode =.. [Name|Args],
  	modeb(Mode),
	get(Const, Type, Body, Args),
  	atomic(Const),
	Type = #_.

cp_args([], []).
cp_args(_.Body, _.Args) :-
	cp_args(Body, Args).


get(Const, Type, Const._, Type._).
get(Const, Type, _.Body, _.Args) :-
	get(Const, Type, Body, Args).

% The following version is marginally faster, but puts the disjuncts first
% contains_constant((H :- T), Const, Type, (H :- Disjunct,NewClause), Disjunct) :-
%    contains_constant_rec(T, Const, NewClause, Var, Type),
%    nonvar(Const),
%    get_new_disjunct(Name),
%    Disjunct =.. [Name, Var],
%    OrigFact =.. [Name, Const],
%    assert(OrigFact).

replace_end_constant((G,T), Const, (G,T2), Var, Type) :- !,
	replace_end_constant(T, Const, T2, Var, Type).
replace_end_constant('!'(G,T), Const, '!'(G,T2), Var, Type) :- !,
	replace_end_constant(T, Const, T2, Var, Type).
replace_end_constant(T, Const, T2, Var, Type) :-
        replace_constant(T, Const, T2, Var, Type).

replace_constant((H,T), Const, (Changed,T2), Var, Type) :- !,
  	(
	  replace_constant(H, Const, Changed, Var, Type), T = T2
	;
	  replace_constant(T, Const, T2, Var, Type), H = Changed
	).
replace_constant((H,T), Const, '!'(Changed,T2), Var, Type) :- !,
  	(
	  replace_constant(H, Const, Changed, Var, Type), T = T2
	;
	  replace_constant(T, Const, T2, Var, Type), H = Changed
	).
replace_constant(H, Const, Changed, Var, Type) :-
  	H =.. [Name|Body],
	cp_args(Body, Args),
  	Mode =.. [Name|Args],
  	modeb(Mode),
	select(Const, Type, Var, Body, Args, NewBody),
  	atomic(Const),
  	\+atom_concat(provar,_,Const),
	Type = #T,
	ok_type(T),
  	%create_new_variable(Type,Var),
  	Changed =.. [Name|NewBody].

select(Const, Type, Var, Const.Body, Type._, Var.Body).
select(Const, Type, Var, A.Body, _.Args, A.NBody) :-
	select(Const, Type, Var, Body, Args, NBody).


