
:-module(vista_demand_constant, [demand_constant/4]).

:- use_module(library(lists), [member/2]).

:- use_module(bottom_clause, [constant_from_bcs/3]).

:- use_module(examples,[example/3]).

:- use_module(library(ordsets),[ord_intersect/3]).

%
% generate a list with all covered examples
%
demand_constant(Cs, _Type, (Ex :- ClauseVar)/C, _C) :-
	eraseall(covered_example),
	eraseall(free_example),
	example(Id,pos,Ex),
	% skeleton covered?
	\+ \+ user:ClauseVar,
	(member(C, Cs),
	    once(user:ClauseVar)
	-> 
	  recordz(covered_example, Id, _)
        ;
	  recordz(free_example, Id, _)
        ),
	fail.
demand_constant(_Cs, Type, (Ex :- ClauseVar)/C, C) :-
	predicate_property(idb:free_example,number_of_clauses(TF)),
	predicate_property(idb:covered_example,number_of_clauses(TC)),
	TF > TC/4,
	% pick a constant, any constant
	vista_eval_node:minpos(N),
	copy_term((Ex :- ClauseVar)/C,(Ex2 :- ClauseVar2)/C),
	N2 is max(N,TC/4),
	constant_from_bcs(Type, N2, C),
	% find two examples that do not cover it.
	once(valid(Ex, ClauseVar, Ex2, ClauseVar2, TF, TC)).

valid(Ex, ClauseVar, Ex2, ClauseVar2, TF, TC) :-
	recorded(free_example,Id1,_),
	example(Id1,pos,Ex),
	once(user:ClauseVar),
%	recorded(free_example,Id2,_),
%	Id2 > Id1,
%	example(Id2,pos,Ex2),
%	once(user:ClauseVar2),
	n_free(Ex2,ClauseVar2, NFree),
	NFree > TC/4,
	n_covered(Ex2,ClauseVar2, NCov),
	( NCov =:= 0 ; NFree/TF > NCov/TC ).


n_covered(Ex, Body, Tot) :-
	Acc = s(0),
	( 
	    recorded(covered_example,Id,_),
	    example(Id, pos, Ex),
	    once(user:Body),
	    Acc = s(I0),
	    I is I0+1,
	    nb_setarg(1,Acc,I),
	    fail
	;
	    Acc = s(Tot)
	).
	
n_free(Ex, Body, Tot) :-
	Acc = s(0),
	( 
	    recorded(free_example,Id,_),
	    example(Id, pos, Ex),
	    once(user:Body),
	    Acc = s(I0),
	    I is I0+1,
	    nb_setarg(1,Acc,I),
	    fail
	;
	    Acc = s(Tot)
	).
	
