
:- module(vista_disjoin,[reduce_disjunction/1]).

:- use_module('../common/debug_write',[debug_format/3]).

:- use_module('../scorer/scorer.yap',[better_score/2,
				      current_test_score/1]).

:- use_module('../vista',[start_new_search/1,
	update_network/1]).

:- use_module(newpred,[get_pred_name/1]).

:- use_module(modes,[ modeb/2,
		      modeb/1,
		      unassert_modeb/1]).

:- use_module(newpred, [assert_extensionally/2]).

:- use_module(variables, [create_new_variable/2]).

:- use_module(library(lists), [append/3]).

reduce_disjunction(OldScore) :-
	R1 is integer(Rule1),
	R2 is integer(Rule2),
	number_atom(R1,AtomNum1),
    	atom_concat('p', AtomNum1, Name1),
	number_atom(R2,AtomNum2),
	atom_concat('p', AtomNum2, Name2),
	% X is integer(2.0).
	%debug_format(1, 'created ~w ~w~n', [Name1,Name2]),
	%spy(disjoin/7),
	disjoin(Name1, Name2, NewHead, Type, InMode1, Clause1, Clause2, VarClause1, VarClause2),
	%debug_format(1, 'did disjoin work ~w ~w ~w ~n', [NewHead, Type, InMode1]),
	debug_format(1, 'Learning disjunction ~w between rule ~w and rule ~w~n', [InMode1, Name1,Name2]),
	debug_format(1, 'scored ~w / oldScore ~w~n', [Score, OldScore]),
	start_new_search(Type),
	!,
	
	(better_score(Score, OldScore) ->
	
		finish_disjoin(VarClause1, VarClause2, InMode1, Score, Name1, Name2, Clause1, Clause2),
		assert(learned(InMode1)),
		debug_format(4, 'completed disjoin~n', [])
		;
		debug_format(4, 'disjoin does not meet threshold~n', []),
		true
		%I am not sure, but should do unassert the facts list below?
		%JJD unassert learned(Name1)
		%JJD unassert learned(Name2) 
	).	
			


finish_disjoin(VarClause1, VarClause2, InMode1, Score, Name1, Name2, Clause1, Clause2) :-
	update_network(Score),
        (search_assert(no_layer) ->
	 true
	 ;
	 %JJD need to unassert modeb for old clauses,
	 VarClause1 = (_ :- B1),
	 VarClause2 = (_ :- B2),
	 debug_format(1, '~w ~w ~w ~w~n', [Name1,Name2,B1,B2]),
	 unassert_modeb( B1 ),
	 unassert_modeb( B2 ),
	 %JJD need to assert modeb for new clause
	 modeb(*,InMode1),
         assert_extensionally(VarClause1, single),
         assert_extensionally(VarClause2, single)
	),
	current_test_score(TestScore),
%	(UseAggregates = single ->
	% MUST UPDATE THEORY!!
	debug_format(1,'~n~n NEW VIEW DISJUNCTION (score ~w/~w):~n    ~w~n    ~w~n~n',[Score,TestScore,Clause1,Clause2]). 
	%assert(learned(PredHead)).
	%type_assert(PredHead).
finish_disjoin(_,_,_).
	

disjoin(Pred1, Pred2, Head, Type, InMode1, Clause1, Clause2) :-
	find_predmode(Pred1, Args),
	get_types(Args, List),
	get_pred_name(PName),
	InMode1 =.. [PName|Args],	
	append([PName], List, HeadList),
	Head =.. HeadList,
	append(['t'], List, TypeList),
	Type =.. TypeList,
	Body1=..[Pred1|List],
%   writeln(Body1),
	Body2=..[Pred2|List],
	Clause1 = (Head :- Body1), 
	Clause2 = (Head :- Body2).

find_learned(Name, Args) :-
	learned(X),
	X =..[Name|Args].

find_predmode(Name, Args) :-
	modeb(X),
	X =..[Name|Args].

get_types([],[]).
get_types([+H|T], Typelist) :-
	get_types(T, List),
	create_new_variable(H, NewVar),
	append([NewVar], List, Typelist).

get_types([-H|T], Typelist) :-
	get_types(T, List),
	create_new_variable(H, NewVar),
	append([NewVar], List, Typelist).
	

