
:- module(vista_eval_node,[evaluate_clause/4,
			   init_eval/0,
			   final_eval_clause/6,
			   store_final_clause/6]).

:- use_module(variabilise,[variabilise/2]).

:- use_module(clause_optimise,[optimise_clause/2]).

:- use_module('../common/debug_write',[debug_format/3]).

:- use_module(valid,[strict_valid_clause/1,
		     valid_clause/1]).

:- use_module('../scorer/scorer.yap',[better_score/2,
	better_entry_score/2,
	increased_score/2]).

:- use_module(search_space,[add_node_to_search_space/2]).

:- use_module(aggregation,[aggregation/3]).

:- use_module(score,[score_clause/6]).

:- use_module(prune,[doprune/1]).

:- use_module(hierarchy,[hier_clause/8,
	commit_hierarchy/2]).

:- dynamic view_temporary/6.

:- dynamic minpos/1.

minpos(1).

:- dynamic minacc/1.

minacc(0.0).

:- dynamic nodes/1.

nodes(1000).

:- dynamic search_mode/1.

search_mode(eager).

:- dynamic dynamic_hierarchy/1.

dynamic_hierarchy(off).

:- dynamic strict_clauses/1.

strict_clauses(off).

init_eval :-
	retractall(view_temporary(_,_,_,_,_,_)).

%
% get clause from IDB and send it back for evaluation.
%
final_eval_clause(Clause,VarClause,AggregatedClause,Type,Score,UseAggregates) :-
	retract(view_temporary(Clause,VarClause,AggregatedClause,Type,Score,UseAggregates)),
	score_clause(AggregatedClause, _, Type, UseAggregates, _TPos, _TNeg).

store_final_clause(GClause, NewClause, AggregatedClause, Type, Score, UseAggregates) :-
	assert(view_temporary(GClause,NewClause,AggregatedClause,Type,Score,UseAggregates)).

% one_more_clause(Clause, _, _, _, _, _) :-
	
% 	add_node_to_search_space(-inf, Clause),
% 	fail.
evaluate_clause(Clause, Type, UseAggregates, OldScores) :-
	get_value(refinements,Rs),
	variabilise(Clause, VarClause),
	validate_clause(VarClause),
	increment_processed_clauses(Cls0),
	stop_search(Cls0,Clause),
	try_to_prune(VarClause, Cls0, Rs),
	( view_temporary(_,_,_,_,BaseScore,_) ->
	  true
	;
	  BaseScore = OldScores
	),
	one_more_clause(Clause, VarClause, UseAggregates, Type, BaseScore, OldScores).

try_to_prune(VarClause, Cls0, Rs) :-
	doprune(VarClause),
	!,
	debug_format(2,'[ ~d/~d  PRUNE ~w !!!! ]~n',[Cls0,Rs,VarClause]),
	fail.
try_to_prune(_, _, _).

one_more_clause(Clause, VarClause, UseAggregates, Type, BaseScores, OldScores) :-
	process_clause(VarClause, AggregatedClause, UseAggregates, on),
	score_clause(AggregatedClause, Scores, Type, UseAggregates, TPos, TNeg),
	% clause scored with success
	debug_show_clause(Clause, TPos, TNeg, UseAggregates, Scores),
	continue_search(Clause, VarClause, AggregatedClause, Type, Scores, BaseScores, OldScores, UseAggregates, TPos-TNeg),
	fail.

process_clause(VarClause0, AggregatedClause, UseAggregates, Check) :-
	(Check = on ->
	  copy_term(VarClause0, VarClause),
	  optimise_clause(VarClause, OptimisedClause)
	% everything should be connected.
%	  OptimisedClause \= ( _ :-  '!'(_,_))
	;
	  VarClause0 = OptimisedClause
	),
	aggregation(OptimisedClause, AggregatedClause, UseAggregates).

continue_search(Clause, VarClause, AggregatedClause, Type, Scores, BaseScores, OldScore, UseAggregates,TPos-TNeg) :-
				%modify for best first search...
	%jjd beam search
	minpos(MinPos),
	TPos >= MinPos,
    assert(score_attributes:score(Scores)),
%	retractall(view_temporary(_,_,_,_,_,_)),
	% always include the node as a node to be explored...
	Prec is TPos/(TPos+TNeg),
	add_node_to_search_space(Prec, Clause),
%	Clause \= (_ :- _,_),
	dynamic_hierarchy(off),
	% check if a score has improved.
	minacc(MinAcc), TPos/(TPos+TNeg) > MinAcc,
	better_score(Scores, OldScore),
	increased_score(Scores, BaseScores), !,
	% commit to this score?
	new_better_score(Scores,Clause,VarClause,AggregatedClause,Type,UseAggregates).
continue_search(Clause, VarClause, _AggregatedClause, Type, Scores, BaseScores, OldScores, UseAggregates, TPos-TNeg) :-
				%modify for best first search...
	dynamic_hierarchy(on),
	minpos(MinPos),
	TPos >= MinPos,
	increased_score(Scores, BaseScores),
	minacc(MinAcc), TPos/(TPos+TNeg) > MinAcc,
	writeln(trying:MinPos:TPos/TNeg:Scores:BaseScores),
	(  VarClause = (_ :- _, _ ), 
	   better_entry_score(Scores, OldScores),
          hier_clause(VarClause, Scores, -1, TPos, UseAggregates, NewScore, NewClause,CType),
	 increased_score(NewScore, Scores)
        ->
	 commit_hierarchy(CType,NewClause)
	;
  TPos > MinPos/2,
	  Scores = NewScore,
	  NewClause = VarClause
	),
	better_score(NewScore, OldScores), !,
	portray_clause(user_error, NewClause),
	process_clause(NewClause, NewAggregatedClause, UseAggregates, off),
	new_better_score(NewScore,Clause,NewClause,NewAggregatedClause,Type,UseAggregates).

new_better_score(Scores,Clause,VarClause,AggregatedClause,Type,UseAggregates) :-
	debug_format(1,'[ got score 2 ~w ]~n',[Scores]),
	retractall(view_temporary(_,_,_,_,_,_)),
	assert(view_temporary(Clause,VarClause,AggregatedClause,Type,Scores,UseAggregates)),
	debug_format(1,'[ Found New Clause ]~n',[]),
	portray_clause(Clause),
	( search_mode(eager) -> throw(limit_exceeded) ; true).

stop_search(Cls0, _Clause) :-
	nodes(MaxNodes),
	Cls0 > MaxNodes, !,
	debug_format(1,'[ **** max node limit exceeded (~d) **** ]~n',[MaxNodes]),
	!,
	throw(limit_exceeded).
stop_search(Cls0, Clause) :-
	Cls0 mod 100 =:= 0, !,
	nodes(Nods),
	variabilise(Clause, VarClause),
	debug_format(1,'/**** done ~d clauses (max ~d), looking at: ~n',[Cls0,Nods]),
	portray_clause(VarClause),
        debug_format(1, '****/~n',[]).
stop_search(_, _).

debug_show_clause(Clause, Cls0, Rs, UseAggregates, Scores) :-
	debug_format(2,'[ ~d/~d ~w ]~n',[Cls0,Rs,Clause,Scores]),
	debug_show_clause(Clause, UseAggregates).

debug_show_clause(Clause, single) :- !,
	debug_format(2,'[ ~w ]~n',[Clause]).
debug_show_clause(Clause, _) :-
	Clause = (Head :- Body), !,
	functor(Head,_,Ar),
	arg(Ar,Head,Var),
	debug_format(2,'[  ~w ]~n',[(Head :- aggregate(Var,count,Body))]).
debug_show_clause(Clause, _) :-
	debug_format(2,'[ ~w ]~n',[Clause]).

validate_clause((H :- B1, Gs)) :- !,
        do_validate_clause((H :- B1, Gs)).
validate_clause(_).

do_validate_clause(VarClause) :-
	strict_clauses(on), !,
	strict_valid_clause(VarClause).
do_validate_clause(VarClause) :-
	valid_clause(VarClause).

increment_processed_clauses(Cls) :-
	get_value(processed_clauses,Cls0),
	Cls is Cls0+1,
	set_value(processed_clauses,Cls).


