
:- module(vista_example,[call_agg/7,
			 call_example/4,
			 call_body/1,
			 example/4,
			 example/3,
			 new_predicate_to_call/1,
			 assert_example/3,
			 count/3,
			 random_example/2,
			 load_examples_file/3,	
			 next_example/2,
			 init_examples/0]).

:- use_module('../scorer/scorer',
	      [covers_example/4]).

:- dynamic example_count/3.

init_examples :-
	abolish(example/4).

load_examples_file(File, Source, Type) :-
	open(File, read, S),
	repeat,
	read(S, Ex),
	( 
	  Ex = end_of_file
	->
	  !,
	  close(S)
	;
	  assert_example(Source, Type, Ex),
	  fail
	).

assert_example(Source,Type,P) :-
	(
	  retract(example_count(Source,Type, Id))
	->
	  true
	;
	  Id = 0
	),
	Id1 is Id+1,
	nb_setval(ex_id, Id1),
	assert(example_count(Source,Type, Id1)),
	assert_static(example(Id1,Source,Type,P)).

example(Id,PosNeg,Ex) :-
	var(Id), !,
	example2(Id,PosNeg,Ex).
example(Id,PosNeg,Ex) :-
	example2(Id,PosNeg,Ex), !.

example2(Id,PosNeg,Ex) :-
	example(Id,train,PosNeg,Ex).
example2(Id,PosNeg,Ex) :-
	example_count(train,PosNeg,Count),
	example(Id0,tune,PosNeg,Ex),	
	Id is Count+Id0.

random_example(Id, PosNeg) :-
	example_count(train,PosNeg, TrPos),
	example_count(tune,PosNeg,  TuPos),
	Id is integer(random*(TrPos+TuPos))+1.

%
% walk over the examples...
%
:- dynamic cur_ex/2.

cur_ex(0, pos).
cur_ex(0, neg).

next_example(Id, PosNeg) :-
	example_count(train,PosNeg, TrPos),
	example_count(tune,PosNeg,  TuPos),
	retract(cur_ex(Id0,PosNeg)),	
	IdC is ((Id0+1) mod (TrPos+TuPos)),
	Id is IdC+1,
	assert(cur_ex(IdC,PosNeg)).	

call_agg(OV,Set,PosNeg,Ex,Linkage,NBody,Count) :-
	example(_,Set,PosNeg,Ex),
	( call_body(Linkage) ->
	    count(OV, NBody, Count)
	;
	    Count = 0
	).

%n_setof(V, Body, Sols) :-
%	findall(V,call_body(Body),Vs),
%	sort(Vs,Ss),
%	length(Ss, Sols).
count(V,Body,_) :-
	eraseall(n_setof),
	call_body(Body),
	recordzifnot(n_setof,V,_),
	fail.
count(_,_,N) :-
	predicate_property(idb:n_setof,number_of_clauses(N)).

call_example(Type,PosNeg,Clause,Obj) :-
	example(Id,Type,PosNeg,Ex),
	once(do_example(Ex,Clause)),
	covers_example(Obj, PosNeg, Type, Id),
	fail. 

do_example(Ex,(Ex :- Body)) :- !,
	nb_setval(ugh, (Ex :- Body)),
	call_body(Body).
do_example(Ex,Ex).

:- dynamic call_body/1.

call_body(umperdick) :-
	nb_getval(ugh, G), writeln(ugh:G), halt.
call_body((B1,B2)) :-
	call_body(B1),
	call_body(B2).
call_body(B=B).
call_body(true).
call_body('!'(G1,G2)) :-
	call_body(G1),
	!,
	call_body(G2).

new_predicate_to_call(EmptyPred) :-
	clause(call_body(EmptyPred),_), !.
new_predicate_to_call(EmptyPred) :-
	assertz((call_body(EmptyPred) :- user:EmptyPred)).

