:- module(vista_init,[view_init/0]).

:- use_module(modes,[determination_arity/1]).

:- use_module(examples,[assert_example/3]).

:- use_module(constants,[init_constants/0]).

:- use_module(score,[assert_counts/4]).

:- use_module('../scorer/scorer',[scorer_init/0,
				  scorer_has_tuneset/0,
				  new_search/1,
				  vals_to_scorable/3]).

:- use_module(examples,[example/4]).

:- use_module('../common/debug_write', [debug_format/3]).

%
% initialise VISTA engine

:- dynamic view_initted/0.

view_init :-
	view_initted, !.
view_init :-
	scorer_init,
	load_examples,
	gen_count_arrays,
	init_constants,
	set_value(total_processed_clauses,0),
	new_search(0),
	assert(view_initted).

load_examples :-
	load_examples(train, pos),
	load_examples(train, neg),
	(
	    scorer_has_tuneset
	 ->
	     load_examples(tune, pos),
	     load_examples(tune, neg)
	 ;
	     true
	),
	load_examples(test, pos),
	load_examples(test, neg).

load_examples(Source, Type) :-
	atomic_concat([Source,'_',Type],Mod),
	once(current_predicate(_,Mod:P)),
	Mod:P,
	assert_example(Source,Type,P),
	fail.
load_examples(_, _).

gen_count_arrays :-
	init_scoring_data(train,pos,TrPCounts),
	init_scoring_data(train,neg,TrNCounts),
	(
	  scorer_has_tuneset
	->
	  init_scoring_data(tune,pos,TuPCounts),
	  init_scoring_data(tune,neg,TuNCounts),
          TrPCountsF = TrPCounts-TuPCounts,
          TrNCountsF = TrNCounts-TuNCounts
        ;
          TrPCountsF = TrPCounts,
          TrNCountsF = TrNCounts
        ),
	init_scoring_data(test,pos,TePCounts),
	init_scoring_data(test,neg,TeNCounts),
	assert_counts(TrPCountsF,TrNCountsF,TePCounts,TeNCounts),
	fail.
gen_count_arrays.

init_scoring_data(Set,Sign,CountObj) :-
	findall(C,(example(_,Set,Sign,_Ex),C=1),Counts),
	length(Counts,LCounts),
	debug_format(3,'gen ~d entries for ~w/~w ~n', [LCounts,Set,Sign]),
	debug_format(3,'gen weightvector ~w~n', [LCounts]),
	vals_to_scorable(Counts, long, CountObj).

