%
% Manage mode and determination statements.
%

:- module(vista_mode,[determination/2,
		      determination/1,
		      determining/2,
		      determination_arity/1,
		      modeh/2,
		      modeh/1,
		      has_modeh/2,
		      has_modeb/2,
		      modeb/2,
		      modeb/1,
		      show_modes/0,
		      find_modeb/1,
		      head_type/2,
		      goal_type/2,
		      goal_mode/2,
		      unassert_modeb/1,
		      unassert_determination/1]).

:- use_module(examples,[new_predicate_to_call/1]).

:- use_module('../common/debug_write',[debug_format/3]).

:- dynamic determination/1,
	determining/2,
	modeh/1,
	modeb/1,
	has_modeh/2,
	has_modeb/2,
	determination_arity/1,
	head_type/2,
	goal_mode/2,
	goal_type/2.

%
% This is an implementation of Aleph modeb.
%
determination(N/A,M/B) :-
	functor(S,N, A),
	determination(S), !,
	functor(D, M, B),
	assert(determining(S,D)).
determination(ToLearn,_) :-
	determination(ToLearn1),
	functor(ToLearn1,N1,A1),
	format(user_error,'Can only learn a single predicate: ~w bs ~w~n',[N1/A1,ToLearn]).
determination(A/N,M/B) :-
	functor(S,A,N),
	assert(determination_arity(N)),
	assert(determination(S)),
	functor(D, M, B),
	assert(determining(S,D)).

modeh(Determinism, PredMode) :-
	functor(PredMode,F,Ar),
	functor(EmptyPred,F,Ar),
	assert(modeh(PredMode)),
	assert(has_modeh(Determinism,PredMode)),
	PredMode =.. [_|ModeArgs],
	types_from_modes(ModeArgs, Types),
	assert(head_type(EmptyPred,Types)),
	type_info(PredMode, EmptyPred).

%modeb(_Determinism, PredMode) :-
%        modeb(_Determinism, PredMode, layer).

modeb(Determinism, PredMode) :-
	functor(PredMode,F,Ar),
	functor(EmptyPred,F,Ar),
	new_predicate_to_call(EmptyPred),
	% for now, I hope this is what Jesse wants.
        %(search_assert(Assert) ->
        % debug_format(1,'**** Not asserting ~w ~w ****~n',[PredMode,Assert]),
	% true
	% ;
        % debug_format(1,'**** Asserting ~w ~w****~n',[PredMode,Assert]),
	% assert(modeb(PredMode))
	%). 
	assert(modeb(PredMode)),
	assert(has_modeb(Determinism,PredMode)),
	type_info(PredMode, EmptyPred).

show_modes :-
	findall(X, find_modeb(X), LitList),
	print_modes(LitList).

print_modes([]).

print_modes([H|T]) :-
    debug_format(1, "modeb: ~w ~n", [H]),
    print_modes(T).
 
find_modeb(X) :-
	modeb(X),
	\+ user:new_predicate(X),
	predicate_property(user:X,number_of_clauses(Cls)),
	Cls >= 1.
/* functor(X,_,2). */

unassert_modeb( H ):-
        functor(H, N, Ar),
        functor(OMode, N, Ar),
	retractall(modeb(OMode)).

unassert_determination( H ):-
        functor(H, N, Ar),
        functor(OMode, N, Ar),
	retractall(determining(_,OMode)).

type_info(PredMode, EmptyPred) :-
	PredMode =.. [_|Modes],
	types_from_modes(Modes, Types),
	(
	    goal_type(EmptyPred, Types)
	->
	    true
	;
	    assert(goal_type(EmptyPred, Types))
	),
	(
	    goal_mode(EmptyPred, Modes)
	->
	    true
	;
	    assert(goal_mode(EmptyPred, Modes))
	).

types_from_modes([], []).
types_from_modes(Mode.Modes, Type.Types) :-
	type_from_mode(Mode, Type),
	types_from_modes(Modes, Types).

type_from_mode(+Type, Type).
type_from_mode(-Type, Type).
type_from_mode(#Type, Type).



