
:- module(vista_variabilise,[variabilise/2,
			     variables/2,
			     unvariabilise/2,
			     replace_var_with_provar/3]).

:- use_module(modes,[find_modeb/1]).

:- use_module('../common/utils',[remove_dups/2]).

:- use_module(variables,[create_new_variable/2]).

variabilise((Head:-Body),(VarHead :- VarBody)) :-
    !,
    mkvars(Head,[], LVars0, VarHead),
    mkvars(Body,LVars0, _, VarBody).

variabilise((Head),(VarHead)) :-
    mkvars(Head,[], _, VarHead).


%mkvars(Head,Lvars0, LVarsF, VarHead),
mkvars(Head, Lvars0, LVarsF, VarHead) :-
    atom(Head), 
    atom_concat('provar', Suffix, Head), !,
    insert_var(Suffix, Lvars0, LVarsF, VarHead).
mkvars(Head, Lvars, Lvars, Head) :-
    atomic(Head), !.
mkvars(S, Lvars0, LVarsF, NS) :-
    S =.. [FunctorName|Args],
    mkvarsl(Args, Lvars0, LVarsF, NArgs),
    NS =.. [FunctorName|NArgs].

mkvarsl([], Lvars, Lvars, []).
mkvarsl([A|Args], Lvars0, LvarsF, [NA|Nargs]) :-
    mkvars(A, Lvars0, LvarsI, NA),
    mkvarsl(Args, LvarsI, LvarsF, Nargs).

insert_var(Atom,[],[Atom-V],V).

insert_var(Atom,[Atom-V|L],[Atom-V|L],V) :- !.

insert_var(Atom,[H|L],[H|NL],V) :-
    insert_var(Atom,L,NL,V).

%gets all variables previously used in clause
variables((Head :- Body), Variables) :-
    literal_variables(Head, Vars1),
    %write(Body),
    %nl,
    % get variables from clause body
    variables1(Body, Vars2),
    append(Vars1, Vars2, Variables1),
    % remove duplicate variables
    remove_dups(Variables1, Variables),
    !.

% finds the variables used in the head of a clause
variables(Head, Variables) :-
    %write('variable bound to head'),
    %nl,
    literal_variables(Head, Vars1),
    remove_dups(Vars1, Variables),
    !.

variables(_, []).

%
% find the variables in the body of a clause
% Lit1 = current literal in body, Rest = rest of the clause body
% Vars = list of variables found in clause body
variables1( (Lit1, Rest), Vars) :-
    !,
    %write('variables1 lit is '),
    %write(Lit1),
    %nl,
    variables1( Lit1, Vars1),
    variables1( Rest, Vars2),
    append( Vars1, Vars2, Vars).

%
% Base case for predicate variables1 - the clause body has only one literal
% Lit1 = current literal in body
% Vars = list of variables found in clause body
% 
variables1(Lit1, Vars) :-
    %write('base of vars1 '),
    %write(Lit1),
    %nl,
    literal_variables(Lit1, Vars).
  
%
% Head = literal to 
% Vars = list of variables used in this literal
%
literal_variables(\+ Head, Vars) :- !,
	literal_variables(Head, Vars).

literal_variables(Head, Vars) :-
    Head =.. [_| Args],
    extract_vars(Args, Vars).

extract_vars([], []).

%
% variables have the following form provar then type then number
%
extract_vars([Head | Tail], [Head | Rest]) :-
	name(Head, [112, 114, 111, 118, 97, 114 | _ ]),
	!,
	extract_vars(Tail, Rest).

extract_vars([_Head | Tail], Rest) :-
    extract_vars(Tail, Rest).

unvariabilise(G, NG) :-
	copy_term(G, NG),
	unvariabilise(NG).

unvariabilise((VarHead:-VarBody)) :- !,
	unvariabilise(VarHead),
	unvariabilise(VarBody).
	
unvariabilise((H , T)) :- !,
	unvariabilise(H),
	unvariabilise(T).
	
unvariabilise(VarHead) :-
	VarHead =.. [_ | Vars],
	unvariabilise2(Vars).
	
unvariabilise2([]).
unvariabilise2([V1 | L1]) :-
	var(V1), !,
	% determine type?
	create_new_variable('type',V1),
	unvariabilise2(L1).
unvariabilise2([_ | L1]) :-
	unvariabilise2(L1).
	
% Replaces all variables in Clause with the provar-ized atom Provar
% and places the result in NewClause
replace_var_with_provar(Clause, Provar, NewClause) :-
	Clause =.. L,
	replace_var_with_provar_list(L, Provar, L2),
	NewClause =.. L2.

replace_var_with_provar_list([], _Provar, []).

replace_var_with_provar_list([H | T], Provar, [H2 | T2]) :-
	compound(H), !,
	replace_var_with_provar(H, Provar, H2),
	replace_var_with_provar_list(T, Provar, T2).

replace_var_with_provar_list([H | T], Provar, [H | T2]) :-
	atomic(H), !,
	replace_var_with_provar_list(T, Provar, T2).
	
replace_var_with_provar_list([H | T], Provar, [Provar | T2]) :-
	var(H),
	replace_var_with_provar_list(T, Provar, T2). 
	
	
